/*  plotdirectionfield.cpp - 
    This file is part of Kalamaris
    Copyright (C) 2001  Antonio Larrosa Jimenez
    Kalamaris' homepage : http://devel-home.kde.org/~larrosa/kalamaris.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "plotdirectionfield.h"
#include "plotorbit.h"
#include "../parameter.h"
#include "../constant.h"
#include "../matrix.h"
#include "../tmatrix.h"
#include "../vartype.h"
#include "../qtai/qtai_plot.h"
#include "graphs.h"
#include <qstring.h>
#include <kdebug.h>
#include <kapp.h>
#include <unistd.h>

PlotDirectionField::PlotDirectionField(void) : MapVariableArgumentsInternal(QString::null,"PlotDirectionField",false)
{
  m_nParameters=4;
  m_parameters=new Parameter *[m_nParameters];

  m_parameters[0]=new Parameter("f");
  m_parameters[0]->setDataType("Matrix");
  m_parameters[1]=new Parameter("bottomleftcorner");
  m_parameters[1]->setDataType("Matrix");
  m_parameters[2]=new Parameter("upperrightcorner");
  m_parameters[2]->setDataType("Matrix");
  m_parameters[3]=new Parameter("partition");
  m_parameters[3]->setDataType("Matrix");
}

PlotDirectionField::PlotDirectionField(const PlotDirectionField &m) : MapVariableArgumentsInternal(m)
{
}

PlotDirectionField::~PlotDirectionField()
{
}
  
Map *PlotDirectionField::copy(void) const
{
  return new PlotDirectionField(*this);
}

Map *PlotDirectionField::eval(const QDict <Map> &vars)
{
  kdDebug() << "PlotDirectionField::eval" << endl;
  QDict <Map> *vars2=evalParams(vars);
  if (!checkParamTypes(*vars2)) return 0L;

  TMatrix &f=*(new TMatrix(*vars2->find("$0")->tmatrix()));
  int id_f=Graphs::add(&f);
  const TMatrix bottomleft(*vars2->find("$1")->tmatrix());
  const TMatrix upperright(*vars2->find("$2")->tmatrix());
  const TMatrix partition(*vars2->find("$3")->tmatrix());
  double xmin=bottomleft.tRef(0);
  double ymin=bottomleft.tRef(1);
  double xmax=upperright.tRef(0);
  double ymax=upperright.tRef(1);
  int xpart=partition.tRef(0);
  int ypart=partition.tRef(1);

  kdDebug() << "vars\n";
  kdDebug() << xmin << " " << ymin << " " << xmax << " " << ymax << "\n";
  kdDebug() << xpart << " " << ypart << "\n";
  
  unsigned int n=f.nRows();
  
  unsigned int i;
  for (i=0;i<n;i++) 
  {
    f.at(i)->propagateForceEvaluation();
  }

  double dx=(xmax-xmin)/xpart;
  double dy=(ymax-ymin)/ypart;
  Map *r;

  
  double *xderiv= new double[xpart*ypart];
  double *yderiv= new double[xpart*ypart];

  Garbage::push();
  for (int i=0;i<ypart;i++)
  {
    Garbage::push();
    vars2->replace("y1",Constant::newConstantG(ymin+i*dy));
    for (int j=0;j<xpart;j++)
    {
        vars2->replace("y0",Constant::newConstantG(xmin+j*dx));
    
      r=f.at(0)->eval(*vars2);
      xderiv[i*xpart+j]=*r->t();
      delete r;
      r=f.at(1)->eval(*vars2);
      yderiv[i*xpart+j]=*r->t();
      delete r;
//      printf(":[%d,%d] %g %g\n",i,j,xderiv[i*xpart+j],yderiv[i*xpart+j]);
    }
  Garbage::pop();
  }

  Garbage::pop();

  int id=0;
  AIPlot *plot=0L;
  if (nArgs()==5) // We want to insert the graph on another plot widget which
    // is already created
  {
    kdDebug() << "Merge plots" << endl;
    T *idx=vars2->find("$4")->t();
    id=*idx;
 
    plot=Graphs::plot2D(id);
    if (plot)
    {
      plot->setDirectionFieldData(xderiv, yderiv, xmin, xmax, ymin, ymax, xpart, ypart);
      xmin=QMIN(xmin, plot->xmin);
      xmax=QMAX(xmax, plot->xmax);
      ymin=QMIN(ymin, plot->ymin);
      ymax=QMAX(ymax, plot->ymax);
      plot->setRanges(xmin,xmax,ymin,ymax);
    } else id=-1;
  }
  else
  {
    plot=new AIPlot(0L,"DirectionField",100,1);
    plot->setDirectionFieldData(xderiv, yderiv, xmin, xmax, ymin, ymax, xpart, ypart);
    double xmin_=QMIN(xmin, plot->xmin);
    double xmax_=QMAX(xmax, plot->xmax);
    double ymin_=QMIN(ymin, plot->ymin);
    double ymax_=QMAX(ymax, plot->ymax);
    kdDebug() << xmin_ << " " << ymin_ << " " << xmax_ << " " << ymax_ << "\n";
    plot->setRanges(xmin_,xmax_,ymin_,ymax_);
    plot->setDoubleBuffer(TRUE);
    plot->setPlotType( AI_PLOT_TREND );
    plot->resize(200,200);
    plot->show();
    id=Graphs::add(plot);   
  }
  if (plot) plot->setCustomMode("Initial Condition Mode",id, id_f,
        static_cast <int (*)(int,int,double,double)>(&PlotOrbit::plotOrbit));
 
  kdDebug() << "PlotDirectionField::eval(end)" << endl;
  delete vars2;
  
  return new Constant(id);
}

QString PlotDirectionField::string(void) const
{
  return "PlotDirectionField";
}

QString PlotDirectionField::help(void) const
{
  return helpPrefix()+"<H3><I>PlotDirectionField( ... )</I></H3>"
    " "+helpPostfix();
}
