/*  kalamaris.cpp - Kalamaris' main widget
    This file is part of Kalamaris
    Copyright (C) 2000  Antonio Larrosa Jimenez
    Kalamaris' homepage : http://devel-home.kde.org/~larrosa/kalamaris.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "kalamaris.h"
#include <math.h>
#include <kfiledialog.h>
#include <kapp.h>
#include <kmenubar.h>
#include <klocale.h>
#include <kdebug.h>
#include <kmessagebox.h>
#include <klineedit.h>
#include <kaction.h>
#include <kiconloader.h>
#include <kstdaction.h>
#include <kio/netaccess.h>
#include <khtml_part.h>
#include <kurl.h>
#include <klineeditdlg.h>
#include <kcombobox.h>
#include <dcopclient.h>
#include <qvbox.h>
#include <qtextview.h>
#include <qsplitter.h>
#include <qscrollbar.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qpushbutton.h>
#include <unistd.h>
#include "map.h"
#include "functions/init.h"
#include "methods/init.h"
#include "methods/mapmethod.h"
#include "internal/init.h"
#include "interpreter.h"
#include "mapbrowser.h"

#include "qtai/qtai_plot.h"

MainWindow::MainWindow ( const char * name ) : KMainWindow ( 0L, name )
{
  interpreter=new Interpreter;

  KStdAction::open(this, SLOT(fileOpen()), actionCollection());
  KStdAction::saveAs(this, SLOT(fileSave()), actionCollection());
  KStdAction::quit(this, SLOT(close()), actionCollection());
  
  KToggleAction *evalfunc=new KToggleAction(i18n("&Evaluate Functions"), BarIcon("funct"), 0, 
      actionCollection(), "flagEvaluateFunctions");
  connect(evalfunc, SIGNAL(toggled(bool)), interpreter, SLOT(setFlagEvaluateFunctions(bool)));
  evalfunc->setChecked(false);

  KSelectAction *availmaps=new KSelectAction(i18n("&Available Maps"), BarIcon("viewmag"), 0, 
			  actionCollection(), "flagAvailableMaps");
  connect(availmaps, SIGNAL(activated(const QString &)), this, SLOT(showHelp(const QString &)));
  
  KSelectAction *datatypes=new KSelectAction(i18n("&Data Type"), 0/*BarIcon("precplus")*/, 0, 
      actionCollection(), "dataType");
  QStringList strlist;
  strlist << "double number" << "arbitrary precision float number" ;
  
  datatypes->setItems(strlist);
  connect(datatypes, SIGNAL(activated(const QString &)), this, SLOT(setDataType(const QString &)));
  
  new KAction(i18n("Set &Precision..."), BarIcon("precplus"), 0, this, SLOT(setPrecision()), 
      actionCollection(), "setPrecision");
  
  KToggleAction *stepcontrol=new KToggleAction(i18n("&Step Control"), BarIcon("undo"), 0,
      actionCollection(), "flagStepControl");
  connect(stepcontrol, SIGNAL(toggled(bool)), interpreter, SLOT(setStepControl(bool)));
  stepcontrol->setChecked(true);

  new KAction(i18n("Examples"), 0, 0, this, SLOT(showExamples()), 
      actionCollection(), "showExamples");

  createGUI("kalamaris.rc");

  QVBox *clientwidget=new QVBox(this);
  splitter=new QSplitter(clientwidget);
  lastcmd=new QTextView(splitter);
  help=new QVBox(splitter);
  QPushButton *helpclosebutton=new QPushButton(i18n("Close Help"),help);
  connect(helpclosebutton,SIGNAL(clicked()), this, SLOT(hideHelp()));

  mapbrowser=new MapBrowser(help);
  connect( mapbrowser, SIGNAL( process( const QString & ) ),
      this, SLOT( helpProcess( const QString & ) ) );

  cmdline=new KHistoryCombo(true,clientwidget,"CommandLine");
  connect(cmdline,SIGNAL(returnPressed(const QString &)),this,SLOT(processCmdline(const QString &)));

  lastcmd->setFocusProxy(cmdline);

  help->hide();

  setCentralWidget( clientwidget );

  connect(interpreter, SIGNAL(outputString( int, const QString &)),
      this, SLOT(interpreterOutput( int, const QString &)));
  connect(interpreter, SIGNAL(inputString( int, const QString &)),
      this, SLOT(interpreterInput( int, const QString &)));

  init_internal();
  init_functions();
  init_methods();

  parseAvailableMaps();

  DCOPClient *client=kapp->dcopClient();
  client->attach();
  client->registerAs("kalamaris", false);
}

void MainWindow::openFile( const QString &filename )
{
  QFile f(filename);
  if (f.open(IO_ReadOnly))
  {
    QTextStream t( &f );
    QString s;
    while ( !t.eof() )
    {
      s= t.readLine();
      cmdline->addToHistory(s);
      interpreter->process(s);
    }
  }
  f.close();
  cmdline->clearEdit();
}

void MainWindow::saveFile( const QString &filename )
{
  QFile f(filename);
  if (f.open(IO_WriteOnly))
  {
    QTextStream t( &f );
    t << history.join("\n");
  }
  f.close();
}

void MainWindow::fileOpen()
{
  KURL url = KFileDialog::getOpenURL( QString::null, "*", this );
  if (url.isEmpty()) return;

  QString tmpFile;
  if( KIO::NetAccess::download( url, tmpFile ) )
  {
    /* Do whatever you want with a _local_ file stored as tmpFile */
    openFile(tmpFile);

    KIO::NetAccess::removeTempFile( tmpFile );
  }	
}

void MainWindow::fileSave()
{
  KURL url = KFileDialog::getSaveURL( QString::null, "*", this );
  if (url.isEmpty()) return;

  char temp[500];
  strcpy(temp,"/tmp/kalamarisXXXXXX");
  mkstemp(temp);
  QString tmpFile=temp;
  saveFile(tmpFile);
  if( KIO::NetAccess::upload( tmpFile, url ) ) unlink(tmpFile);
}

void MainWindow::processCmdline(const QString &text)
{
  kdDebug() << "--------" << endl;
  /*int id=*/interpreter->process(text);
  cmdline->addToHistory(text);
  cmdline->clearEdit();
  cmdline->reset();
}

void MainWindow::parseAvailableMaps()
{
  QStringList names=interpreter->mapNames();
  names.sort();
  KSelectAction *mapsavail=(KSelectAction*)actionCollection()->action("flagAvailableMaps");
  mapsavail->clear();
  mapsavail->setItems(names);
}

void MainWindow::interpreterInput(int id, const QString &input)
{
  lastcmd->append(QString("%1 -- %2").arg(id).arg(input));
  history << input;
  saveFile(".#kalamaris.lastcmds");
  /*
   * This doesn't seem to work
   *
   *  QScrollBar *vsb=lastcmd->verticalScrollBar();
   *  vsb->setValue(vsb->maxValue());
   */
}

void MainWindow::interpreterOutput(int id, const QString &output)
{
  lastcmd->append(QString("%1 >> %2").arg(id).arg(output));
  parseAvailableMaps();
  /*
   * This doesn't seem to work
   *
   *  QScrollBar *vsb=lastcmd->verticalScrollBar();
   *  vsb->setValue(vsb->maxValue());
   */
}

void MainWindow::showHelp(const QString &mapname)
{
  help->show();
  mapbrowser->helpOpenURL(QString("map:%1").arg(mapname));
}

void MainWindow::hideHelp()
{
  help->hide();
}

void MainWindow::helpProcess(const QString &cmd)
{
  interpreter->process(cmd);
}

void MainWindow::setDataType(const QString &t)
{
  if (t=="double number") T::setDataType(T::T_);
  else T::setDataType(T::TArbitraryPrecision_);
  // XXX We should call each Constant object and tell it to convert its data
  // types or implement methods that operate on both of them at the same time
  // (probably much better)
}

void MainWindow::setPrecision()
{
  bool ok;
  QString r=KLineEditDlg::getText(
      i18n("Enter number of decimal numbers to display"),
      QString::number(T::decimalNumbers()),&ok,0);

  if (ok)
  {
    kdDebug() << r.toInt() << endl;
    T::setPrintDecimalNumbers(r.toInt());
  }
}

void MainWindow::showExamples()
{
  help->show();
  mapbrowser->helpOpenURL("demo://index");
}
