/*  kglarrow.cpp - 
    This file is part of KGL
    Copyright (C) 2000,2001  Antonio Larrosa Jimenez
    KGL' homepage : http://devel-home.kde.org/~larrosa/kgl.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "kglarrow.h"
#include <math.h>
#include <stdio.h>

KGLArrow::KGLArrow(GLfloat x1, GLfloat y1, GLfloat z1, GLfloat x2, 
    GLfloat y2, GLfloat z2, GLfloat r, GLfloat r2,
    int slices, QImage texture1, QImage texture2) : KGLObject(x1,y1,z1)
{
  GLfloat h=sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)+(z2-z1)*(z2-z1));
//  printf("h %g\n",h);
  
/* Let's calculate the angle between (0,0,0)(0,0,h) and (0,0,0)(x2-x1,y2-y1,z2-z1)
 * We need first to get the orthogonal projection of (0,0,h) onto the line 
 * generated by the vector (x2-x1,y2-y1,z2-z1) 
 *
 * We have to get a point (x,y,z) which verifies:
 *  x-x1   y-y2    z-z2
 *  ---- = ----  = ----
 *  x2-x1  y2-y1   z2-z1
 *
 * and
 *
 * (a-x)*(x1-x)+(b-y)*(y1-y)+(c-z)*(z1-z)=0
 *
 * That is, (x,y,z) which verifies :
 *
 *    x      y
 *  ---- = ----
 *  x2-x1  y2-y1
 *
 *    y       z
 *  ----  = ----
 *  y2-y1   z2-z1
 *
 * and
 *
 *  2   2
 * x  +y  +z*(z-h)=0
 *
 * Therefore, we must have:
 *
 *             (z2-z1)*h*(y2-y1)
 * y =  -------------------------------------
 *                2           2           2
 *         (x2-x1)   + (y2-y1)   + (z2-z1)
 *
 * and:
 *      y*(x2-x1)       y*(z2-z1)
 *  x = ---------   z = ---------
 *        y2-y1           y2-y1
 */

 GLfloat dx=x2-x1;
 GLfloat dy=y2-y1;
 GLfloat dz=z2-z1;

/**
  * Even faster, the dot product of a and b is the scalar number by which we 
  * have to multiply b to get the orthogonal projection of a over b
  */

 GLfloat dp=h*dz/(dx*dx+dy*dy+dz*dz);
 GLfloat px=dx*dp;
 GLfloat py=dy*dp;
 GLfloat pz=dz*dp;
 
// printf("p %g\t%g\t%g\n",px,py,pz);

 GLfloat dpx=sqrt(px*px+py*py+pz*pz);
 GLfloat dpy=sqrt(px*px+py*py+(h-pz)*(h-pz));

// printf("dp %g\t%g\n",dpx,dpy);
 
// rho is the angle between (0,0,h) and (dx,dy,dz)
 GLfloat rho=atan2(dpy,dpx);
 if (dp<0.0) rho=M_PI-rho;
 
// printf("rho %g\n",rho*360.0/(2*M_PI));
// v is the vectorial product of (0,0,h) and d
 GLfloat vx=-h*dy;
 GLfloat vy=h*dx;
 GLfloat vz=0;
 
// printf("v %g\t%g\t%g\n",vx,vy,vz);

  m_list = glGenLists(1);
  glNewList(m_list, GL_COMPILE);
  GLUquadricObj* q = gluNewQuadric();
  gluQuadricDrawStyle( q, (GLenum) GLU_FILL );
  gluQuadricNormals( q, (GLenum) GLU_SMOOTH );

  glPushMatrix();
  glMatrixMode(GL_MODELVIEW);

  if (!texture2.isNull()) 
  {
    gluQuadricTexture(q, GL_TRUE);
    texture2 = texture2.convertDepth( 32 ).mirror().swapRGB(); 
    glTexImage2D( GL_TEXTURE_2D, 0, 3, texture2.width(), texture2.height(), 0,
	GL_RGBA, GL_UNSIGNED_BYTE, texture2.bits() );
  }

  gluQuadricNormals(q, GL_SMOOTH);
  glRotatef(rho*360.0/(2*M_PI),vx,vy,vz);
  gluCylinder(q, r, r*0.95, h*3/4, slices, 1);

  // end cap

  if (!texture1.isNull()) 
  {
    gluQuadricTexture(q, GL_TRUE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, glcolor(Qt::white) );
    texture1 = texture1.convertDepth( 32 ).mirror().swapRGB(); 
    glTexImage2D( GL_TEXTURE_2D, 0, 3, texture1.width(), texture1.height(), 0,
	GL_RGBA, GL_UNSIGNED_BYTE, texture1.bits() );
  } else {
    gluQuadricTexture(q, GL_FALSE);
  }
  glTranslatef(0.0, 0.0, h*3/4);
  gluCylinder(q, r2, 0.0, h/4, slices, 1);

  // other end cap

  glTranslatef(0.0, 0.0, -h*3/4);

  gluQuadricOrientation(q, (GLenum)GLU_INSIDE);
  gluDisk(q, 0.0, r, slices, 1);
  gluQuadricOrientation(q, (GLenum)GLU_OUTSIDE);
  glRotatef(-rho,vx,vy,vz);
  glPopMatrix();
  glEndList();

  gluDeleteQuadric(q);
}
