/*  map.cpp - 
    This file is part of Kalamaris
    Copyright (C) 2000  Antonio Larrosa Jimenez
    Kalamaris' homepage : http://devel-home.kde.org/~larrosa/kalamaris.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "map.h"
#include <kdebug.h>
#include <klocale.h>
#include <qregexp.h>
#include "operator.h"
#include "constant.h"
#include "variable.h"
#include "matrix.h"
#include "mextract.h"
#include "parameter.h"
#include "internal/derive.h"

Map::Map (const QString &decl,const QStringList &parms, const QString &nam, bool parsedefinition)
{
  m_isAClassName="Map";
  m_tree=0L;
  m_definition=decl;
  m_forceEvaluation=false;
  m_nParameters=parms.count();

  m_parameterList=new QStringList(parms);
  m_parameters=new Parameter *[m_nParameters];
  
  for (unsigned int i=0;i<m_nParameters;i++)
  {
    m_parameters[i]=new Parameter(*parms.at(i),*parms.at(i));
  }

  setName(nam);

  if (m_definition.isEmpty()) return;

  if (parsedefinition) m_tree=parse(m_definition);
}

Map::Map(const Map &m)
{
  m_isAClassName="Map"; // Note that (e.g.) even if we create a new Map from 
                        // a Constant, it's a map anyway
  if (m.m_tree!=0L) m_tree=m.m_tree->copy();
  else m_tree=0L;
  m_definition=m.m_definition;
  m_forceEvaluation=m.m_forceEvaluation;
  m_nParameters=m.m_nParameters;
  m_parameterList=new QStringList(*m.m_parameterList);
  m_parameters=new Parameter *[m_nParameters];

  for (unsigned int i=0;i<m_nParameters;i++)
  {
    m_parameters[i]=m.m_parameters[i]->copy();
  }
  m_name=m.m_name;
}

Map::Map (const QString &nam, int)
{
  m_isAClassName="Map";
  m_tree=0L;

  m_parameterList=0L;
  m_parameters=0L;
  m_nParameters=0;
  m_forceEvaluation=false;

  setName(nam);
}

Map::~Map()
{
  delete m_tree;

  for (unsigned int i=0;i<m_nParameters;i++)
    delete m_parameters[i]; 

  if (m_parameters) delete [] m_parameters;
  delete m_parameterList;
}

void Map::setName(const QString &name, bool insertInDict)
{
  m_name=name.stripWhiteSpace();
  if ((!m_name.isEmpty())&&(insertInDict)) 
  {
    maps.insert(m_name,this);
    kdDebug() << "  maps dictionary += ----------------------->  " << m_name << endl;
  }
  
}

void Map::setParameterExpression(unsigned int i,const QString &expr)
{
  if (i>=m_nParameters) 
  {
    kdError() << QString("ERROR !!!! : setParameterExpression : %1 > %2 (%3)")			.arg(i).arg(m_nParameters).arg(expr) << endl;
    printDebug();
    return;
  }
  QString pname=m_parameters[i]->name();
  delete m_parameters[i];
//  kdDebug() << QString("Map::setParameterExpression (%1) <- (%2)")//      				.arg(pname).arg(expr) << endl;
  m_parameters[i]=new Parameter(pname,expr);
}

void Map::setParameterExpression(unsigned int i, Map *expr)
{
  if (i>=m_nParameters) kdError() << "ERROR !!!! : setParameterExpression : " << i << " > " << m_nParameters << endl;

  QString pname(QString("x%1").arg(i));
  if (m_parameters[i]) 
  {
    pname=m_parameters[i]->name();
    delete m_parameters[i];
  }
//  kdDebug() << QString("Map::setParameterExpression (%1) (%2)").arg(pname)// 					.arg(expr->string()) << endl;
  m_parameters[i]=new Parameter(pname,expr);
}

void Map::setParameterExpression(unsigned int i, Parameter *expr)
{
  if (i>=m_nParameters) 
  {
    kdError() << QString("ERROR !!!! : setParameterExpression : %1 > %2")	.arg(i).arg(m_nParameters) << endl;
    printDebug();
    return;
  }

  if (m_parameters[i]) delete m_parameters[i];
  m_parameters[i]=expr;
}

QDict <Map> *Map::evalParams(const QDict <Map> &vars)
{
  QDict <Map> *vars2= new QDict <Map>(vars);
  Map *t;
  for (unsigned int i=0;i<m_nParameters;i++)
  {
    t=m_parameters[i]->eval(vars);
//    kdDebug() << QString("t (%1)<<(%2)").arg(params[i]->name()).arg(t->string()) << endl;
    
    // X_X If we replace it, we're leaking memory !
    //  : Not if we use the new Garbage collector !
    Garbage::throwMap(t);
    vars2->replace(m_parameters[i]->name(),t);
  }

  return vars2;
}


Map *Map::eval(const QDict <Map> &vars)
{
//  kdDebug() << QString("Map::eval (%1) (%2)").arg(name).arg(string()) << endl;

  QDict <Map> *vars2=evalParams(vars);
  if (m_tree==0L)
    kdError() << "Map::eval with tree==0L, this should never happen," 
      " so I'll leave it crash (" << isA() << ")" << endl;
  
  Map *r=m_tree->eval(*vars2);
  /*
   * Idea: We could as well do a timeout if it takes too long to evaluate
   * the tree, and just return a copy of *this, in that case.
   * Hmm, this may break complex methods which must take long times
   * to compute.
   */

  if (!r->isConstant()&&(!m_name.isEmpty())&&
      (vars.find("$$__noEvaluateFunctions")!=0L)&&(!m_forceEvaluation))
  {
    delete r;
    r=copy();
    for (unsigned int i=0;i<m_nParameters;i++)
      r->setParameterExpression(i,vars2->find(m_parameters[i]->name())->copy());
  }
    
  delete vars2;
  return r;
}

Map *Map::copy(void) const
{
  return new Map(*this);
}

void Map::printDebug(const QString &tab) const
{
  kdDebug() << tab+m_name+" ("+isA()+")" << endl;
  kdDebug() << tab+string() << endl;
  for (unsigned int i=0;i<m_nParameters;i++)
  {
    kdDebug() << tab+QString("param %1").arg(i) << endl;
    m_parameters[i]->printDebug(tab+"      ");
  }

  if (m_tree) m_tree->printDebug(tab+"   ");
  else kdDebug() << tab+"no sub-tree" << endl;
  kdDebug() << tab+"---" << endl;

}
  

Map *Map::parse(QString def)
{
  def=def.simplifyWhiteSpace();
  
  if (def.isEmpty()) return 0L;
  kdDebug() << QString("parse (%1)").arg(def) << endl;

  const QString op("+-*/^");
  const QChar *c=op.unicode();  

  const QChar *p;
  unsigned int bracketdepth,squarebracketdepth;
  int i;
  unsigned int lc=op.length();
  unsigned int lp;
 
  while (lc)
  {
    p=def.unicode();
    lp=def.length();
    i=0,bracketdepth=0,squarebracketdepth=0;
    while (lp)
    {
      if ((*p==*c)&&(bracketdepth+squarebracketdepth==0)) 
      {
	if ((*c!='-')||((i>0)&&(*(c-1)!='*')&&(*(c-1)!='/')&&(*(c-1)!='^'))) 
	{                     // With this condition, we allow negative signs
	  kdDebug() << "Operator" << endl;
	  return new Operator((char)*c,def.left(i),def.right(def.length()-i-1));
	}
      }
      if (*p=='(') bracketdepth++;
      if (*p==')') bracketdepth--;
      if (*p=='[') squarebracketdepth++;
      if (*p==']') squarebracketdepth--;
      // TODO Check for negative values, as that would mean there's a syntax error
      p++; lp--;
      i++;    
    }
    c++; lc--;
  }

  /* If there's no operator, it's because we have :
    1) an extraction from a matrix    "x[3]", "[1,0;0,1][1]"
    2) a parenthesis    "(2x+3)"                  [ 1 ]
    3) a matrix/vector, "[1,0;0,1]", "[1;2;3]" == [ 2 ] or "[1,2,3]" == [ 1  2  3 ]
    4) a number,        "-24.6"                   [ 3 ]
    5) a variable       "x"
    6) a function       "cos(3x)"
  */
  
  kdDebug() << "something" << endl;

#ifndef DONT_ALLOW_FACTORIAL
  if ((def.length()>1)&&(def[def.length()-1]=='!'))
  {
    Map *m=functionFactory("Factorial");
    if (m) m->setParameters(def.left(def.length()-1));    
    return m;
  }
#endif
 
//  QRegExp r  = QRegExp("^-\?[0-9]*$", TRUE, FALSE); 
  QRegExp constantRegExp = QRegExp("^-\?[0-9]*\\.\?[0-9]+", TRUE, FALSE); 
  int len;
//  if ((def.find(r)!=-1)) // It's a constant
  if ((constantRegExp.match(def,0,&len)!=-1)) // It may be a constant
  {					
    if (def.length()!=(unsigned)len) // It's an implicit multiplication,
   			// something of the form "2x" or "4 x" or even "-x"
    {
      kdDebug() << QString("Implicit Multiplication (%1) %2 %3").arg(def).arg(def.length()).arg(len) << endl;
      return new Operator('*',def.left(len),def.mid(len));
    }
    else
    {
    
    // It's a constant of the form 53, -24. or 15.17
     kdDebug() << QString("constant %1").arg(def) << endl;
     return new Constant(def);
    }
  }

  if (def[0]=='-')
  {
    // It's something like "-y" or "-f(x)"
    kdDebug() << QString("negation %1").arg(def) << endl;
    Map *r=parse(def.mid(1));
    return new Operator('*',new Constant(-1.0),(r)? r : new Constant(0.0));
    // NOTE: Perhaps it's better to use a Negate Map-based class ?
  }
  

  if ((def.length()>3)&&(def[def.length()-1]==']')) // It _could_be_ a matrix extraction 
  {
    kdDebug() << QString("possible matrixextract %1").arg(def) << endl;
    lp=def.length()-2;
    p=def.unicode()+lp;
    squarebracketdepth=1;
    bracketdepth=0;

    while ((lp)&&(bracketdepth+squarebracketdepth>0))
    {
      if (*p=='(') bracketdepth--;
      if (*p==')') bracketdepth++;
      if (*p=='[') squarebracketdepth--;
      if (*p==']') squarebracketdepth++;
      // TODO Check for negative values, as that would mean there's a syntax error
      
      p--; lp--;
    }

    if ( ((lp>0) && (constantRegExp.match(def.left(lp+1),0,&len)!=-1)) || ((lp==0) && (def[0]=='-')))
    {
      kdDebug() << QString("Implicit Multiplication by matrix (%1) %2 %3").arg(def).arg(def.length()).arg(len) << endl;
      if (lp==0) return new Operator('*',new Constant(-1.0),parse(def.mid(1)));
      else return new Operator('*',def.left(len),def.mid(len));
      
    }
    else
    if ((lp>0)||((lp==0)&&(*p!='[')&&(bracketdepth+squarebracketdepth==0))) 
    {  					// It _is_ a matrix extraction
      kdDebug() << QString("matrixextract %1").arg(def) << endl;
      /**
       * If we don't check it like we've just done, matrix extractions
       * and plain matrices could be miscast
       */
      lp+=2;
      Map *m=new MatrixExtract(def.left(lp-1));

      m->setParameters(def.mid(lp,def.length()-lp-1));

      return m;
    }
  }

 
  if (def[0]=='(') // It's a parenthesis
  {
    kdDebug() << QString("bracket %1").arg(def) << endl;
    return parse(def.mid(1,def.length()-2));
  }

  if (def[0]=='[') // It's a matrix ( or a vector )
  {
    kdDebug() << QString("squarebracket %1").arg(def) << endl;
    return new Matrix(def.mid(1,def.length()-2));
  }
   
  if (def[0]=='"') // It's a string
  {
    kdDebug() << QString("variable %1").arg(def) << endl;
    return new Variable(def.mid(1,def.length()-2));
  }

  if ((i=def.find('('))!=-1) // It's a function call
  {
     kdDebug() << QString("function %1").arg(def) << endl;
     Map *m;
     // We're no more constructing functions when parsing, but doing
     // lazy evaluation (except for internal functions :) )
     if (isInternal(def.left(i)))
     {
       m=functionFactory(def.left(i));
     }
     else
     {
       m=new MapVariableArguments();
       m->setName(def.left(i),false);
     }

     i++;
    
     kdDebug() << QString("param (%1)").arg(def.mid(i,def.length()-i-1)) << endl;
     if (!m)
      { // This function doesn't exist !!
	// Let's do some magic here
        m=new Variable(def.left(i-1));
      }
	
     if (m) m->setParameters(def.mid(i,def.length()-i-1));
     
     return m;
  }
  else  // It's a variable
  {
    
    kdDebug() << QString("variable %1").arg(def) << endl;
    return new Variable(def);
  }
    
}

void Map::setParameters(const QString &par)
{
  const QChar *p=par.unicode();
  unsigned int i=0,j=0;
  unsigned int lp=par.length(); 
  unsigned int c=0,bracketdepth=0,squarebracketdepth=0;

  while (lp)
  {
    if ((*p==',')&&(bracketdepth+squarebracketdepth==0))
    {
      setParameterExpression(c,par.mid(i,j-i));
      c++;
      i=j+1;
    }
    if (*p=='(') bracketdepth++;
    if (*p==')') bracketdepth--;
    if (*p=='[') squarebracketdepth++;
    if (*p==']') squarebracketdepth--;
    p++; lp--;
    j++;
  }

  setParameterExpression(c,par.mid(i,j-i));
}

int Map::checkParamTypes(const QDict <Map> &vars) const
{
  for (unsigned int i=0;i<m_nParameters;i++) 
    if (!m_parameters[i]->checkDataType(vars)) return false;
  return true;
}

bool Map::isInternal(const QString &name)
{
  Map *m=maps.find(name);

  if (m) return m->isInternal(); 
  return false;
}

Map *Map::functionFactory(const QString &name)
{
//  kdDebug() << QString("Factory (%1)").arg(name) << endl;
  Map *m=maps.find(name);
  Map *r=0L;
  if (m) r=m->copy();

  return r;
}

Map *Map::argument(unsigned int i) const
{
  if (i>=m_nParameters) return 0L;
  return m_parameters[i]->map();
} 

QStringList Map::split(const QChar &sep,const QString &str)
{
  QStringList lst;
  const QChar *c=str.unicode();
  static QChar openbracket('(');
  static QChar closebracket(')');
  static QChar opensqbracket('[');
  static QChar closesqbracket(']');
  unsigned int i=0,j=0;
  unsigned int bracketdepth=0,sqbracketdepth=0;

  while (j<str.length())
  {
    if ((*c==sep)&&(bracketdepth+sqbracketdepth==0))
    {
      lst << str.mid(i, j-i);
      i=j+1;
    }
    if (*c==openbracket) bracketdepth++;
    if (*c==closebracket) bracketdepth--;
    if (*c==opensqbracket) sqbracketdepth++;
    if (*c==closesqbracket) sqbracketdepth--;
    c++;
    j++;
  }
  lst << str.mid(i, j-i);

  return lst;

} 

/*
void Map::avoidNameLookup( const QString &name )
{
  if (!nosubs)
  {
    nosubs=new MapVariableArguments();
    nosubs->setName("$$__noSubstitute",false);
    vars2.insert(QString("$$__noSubstitute"),nosubs);
  }
  else
  {
    if (nosubs->isArgumentName(name))
    {
      kdError() << QString("Recursive Variable evaluation (%1)!!").arg(name) << endl;
      return copy();
    }
  }
}
*/

QString Map::string(void) const
{
  if (!m_name.isEmpty()) 
  {
    QString s(m_name+"( ");

    for (unsigned int i=0;i<m_nParameters;i++)
    {
      s+=m_parameters[i]->string();
      if (i+1<m_nParameters) s+=", ";
    }
    s+=" )";
    
    return s;
  }
  else
  {
    if (m_tree) return QString("\\(%1\\)").arg(m_tree->string());
    else kdError() << "ERROR : ***** Map::string called with no name and no tree ********************** " << endl;
  }
//    return QString("\\(%1\\)").arg(tree->string());

  return "";
}

QString Map::htmlString(void) const
{
  return string();
}

QString Map::helpPrefix(void)
{
  return QString("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\n"
           "\"http://www.w3.org/TR/REC-html40/loose.dtd\">\n<HTML><BODY bgcolor=\"#E6F082\">\n");
}

QString Map::helpPostfix(void)
{
  return QString("\n</BODY></HTML>\n");
}

QString Map::help(void) const
{
  kdDebug() << helpPrefix() << endl;
  if (!m_name.isEmpty()) 
    return helpPrefix()+QString("<H3><I>%1 = %2</I></H3><P>User defined function<P>"
		"<H5><A HREF=\"math://RequestPlotFunction/%3\">Plot Function ...</A></H5>")
      		.arg(string()).arg((m_tree)?(m_tree->string()):(QString("<not defined>"))).arg(m_name)+helpPostfix();
  return helpPrefix()+QString(i18n("No help available"))+helpPostfix();
}

Map *operator +(const Map &m1, const Map &m2)
{
   Map *mm1=const_cast<Map *>(&m1);
   Map *mm2=const_cast<Map *>(&m2);
   return Map::add(mm1,mm2,false,false);
}

Map *Map::add(Map * &m1, Map * &m2, bool managem1, bool managem2)
{
  Map *m=0L;
  
  if ((m1->isA("Constant"))&&(m2->isA("Constant")))
  {
    Constant *c1=(Constant *)m1;
    Constant *c2=(Constant *)m2;
    m=Constant::add(c1,c2);
    if (managem1) { delete m1; m1=0L; }
    if (managem2) { delete m2; m2=0L; }
  }
  else
  {
    T &zero=T::newTG(0.0);
    if ((m1->isA("Constant"))&&(*m1->t()==zero))
    {
      if (managem1) { delete m1; m1=0L; }
      if (managem2) { m=m2; m2=0L; }
      else m=m2->copy();
    }
    else if ((m2->isA("Constant"))&&(*m2->t()==zero))
    {
      if (managem2) { delete m2; m2=0L; }
      if (managem1) { m=m1; m1=0L; }
      else m=m1->copy();
    }
    else
    {
      if ((m1->isA("Matrix"))&&(m2->isA("Matrix")))
      {
	Matrix *mm1=(Matrix *)m1;
	Matrix *mm2=(Matrix *)m2;
	m=Matrix::add(mm1,mm2,managem1,managem2);
      }
      else
      {
	Map *mm1=(managem1)?m1:m1->copy();
	Map *mm2=(managem2)?m2:m2->copy();
	m=new Operator('+', mm1, mm2);
      }
      if (managem1) m1=0L;
      if (managem2) m2=0L;
    }
  }
  return m;
}
Map *operator -(const Map &m1, const Map &m2)
{
   Map *mm1=(Map *)&m1;
   Map *mm2=(Map *)&m2;
   return Map::sub(mm1,mm2,false,false);
}

Map *Map::sub(Map * &m1, Map * &m2, bool managem1, bool managem2)
{
  Map *m=0L;
  
  if ((m1->isA("Constant"))&&(m2->isA("Constant")))
  {
    Constant *c1=(Constant *)m1;
    Constant *c2=(Constant *)m2;
    m=Constant::sub(c1,c2);
    if (managem1) { delete m1; m1=0L; }
    if (managem2) { delete m2; m2=0L; }
  }
  else
  {
    T &zero=T::newTG(0.0);
    if ((m1->isA("Constant"))&&(*m1->t()==zero)) // 0 - m2
    {
      if (managem1) { delete m1; m1=0L; }
      return Map::mul2(new Constant(-1.0), m2, true, managem2 );
      if (managem2) m2=0L;
    }
    else if ((m2->isA("Constant"))&&(*m2->t()==zero)) // m1 - 0
    {
      if (managem2) { delete m2; m2=0L; }
      if (managem1) { m=m1; m1=0L; }
      else m=m1->copy();
    }
    else
    {
      if ((m1->isA("Matrix"))&&(m2->isA("Matrix")))
      {
	Matrix *mm1=(Matrix *)m1;
	Matrix *mm2=(Matrix *)m2;
	m=Matrix::sub(mm1,mm2,managem1,managem2);
      }
      else
      {
	Map *mm1=(managem1)?m1:m1->copy();
	Map *mm2=(managem2)?m2:m2->copy();
	m=new Operator('-', mm1, mm2);
      }
      if (managem1) m1=0L;
      if (managem2) m2=0L;
    }
  }
  return m;
}

Map *operator *(const Map &m1, const Map &m2)
{
  Map *mm1=(Map *)&m1;
  Map *mm2=(Map *)&m2;
  return Map::mul(mm1,mm2,false,false);
}

Map *Map::mul(Map * &m1, Map * &m2, bool managem1, bool managem2)
{
  Map *m=0L;
  
  if ((m1->isA("Constant"))&&(m2->isA("Constant")))
  {
    Constant *c1=(Constant *)m1;
    Constant *c2=(Constant *)m2;
    m=Constant::mul(c1,c2);
    if (managem1) { delete m1; m1=0L; }
    if (managem2) { delete m2; m2=0L; }
  }
  else
  {
    T &zero=T::newTG(0.0);
    T &one=T::newTG(1.0);
    if ( ( m1->isA("Constant") && *m1->t()==zero ) 
	|| ( m2->isA("Constant") && *m2->t()==zero ) )
    {
      if (managem1) { delete m1; m1=0L; }
      if (managem2) { delete m2; m2=0L; }
      m=new Constant(0.0);
    }
    else if ( m1->isA("Constant") && *m1->t()==one ) 
    {
      if (managem1) { delete m1; m1=0L; }
      if (managem2) { m=m2; m2=0L; }
      else m=m2->copy();
    }
    else if ( m2->isA("Constant") && *m2->t()==one ) 
    {
      if (managem2) { delete m2; m2=0L; }
      if (managem1) { m=m1; m1=0L; }
      else m=m1->copy();
    }
    else   
    {
      if ((m1->isA("Matrix"))&&(m2->isA("Matrix")))
      {
	Matrix *mm1=(Matrix *)m1;
	Matrix *mm2=(Matrix *)m2;
	m=Matrix::mul(mm1,mm2,managem1,managem2);
      }
      else
	if  ((m1->isA("Constant"))&&(m2->isA("Matrix")))
	{ 
	  Constant *cc1=(Constant *)m1;
	  Matrix *mm2=(Matrix *)m2;
	  m=Matrix::mul(cc1,mm2,managem1,managem2);
	}
	else if ((m1->isA("Matrix"))&&(m2->isA("Constant"))) 
	{
	  Matrix *mm1=(Matrix *)m1;
	  Constant *cc2=(Constant *)m2;
	  m=Matrix::mul(cc2,mm1,managem2,managem1); //It's commutative, isn't it ?
	}
	else
	{
	  Map *mm1=(managem1)?m1:m1->copy();
	  Map *mm2=(managem2)?m2:m2->copy();
	  m=new Operator('*', mm1, mm2);
	}
      if (managem1) m1=0L;
      if (managem2) m2=0L;
    }
  }
  return m;
}

Map *operator /(const Map &m1, const Map &m2)
{
  Map *mm1=(Map *)&m1;
  Map *mm2=(Map *)&m2;
  return Map::div(mm1,mm2,false,false);
}

Map *Map::div(Map * &m1, Map * &m2, bool managem1, bool managem2)
{
  Map *m=0L;
  
  if ((m1->isA("Constant"))&&(m2->isA("Constant")))
  {
    Constant *c1=(Constant *)m1;
    Constant *c2=(Constant *)m2;
    m=Constant::div(c1,c2);
    if (managem1) { delete m1; m1=0L; }
    if (managem2) { delete m2; m2=0L; }
  }
  else
  {
    if ((m1->isA("Matrix"))&&(m2->isA("Matrix")))
    {
      Matrix *mm1=(Matrix *)m1;
      Matrix *mm2=(Matrix *)m2;
      m=Matrix::div(mm1,mm2,managem1,managem2);
    }
    else
    {
      Map *mm1=(managem1)?m1:m1->copy();
      Map *mm2=(managem2)?m2:m2->copy();
      m=new Operator('/', mm1, mm2);
    }
    if (managem1) m1=0L;
    if (managem2) m2=0L;
  }
  return m;
}

Map *Map::pow(Map * &m1, Map * &m2, bool managem1, bool managem2)
{
  Map *m=0L;
  
  if ((m1->isA("Constant"))&&(m2->isA("Constant")))
  {
    Constant *c1=(Constant *)m1;
    Constant *c2=(Constant *)m2;
    m=Constant::pow(c1,c2,managem1,managem2);
  }
  else 
  {
    if ( m2->isA("Constant") && (*m2->t()==T::newTG(0.0)) )
    {
      if (managem1) { delete m1; m1=0L; }
      if (managem2) { delete m2; m2=0L; }
      return new Constant(1.0);
    }
    else if ( m2->isA("Constant") && (*m2->t()==T::newTG(1.0)) )
    {
      if (managem2) { delete m2; m2=0L; }
      if (managem1) { m=m1; m1=0L; }
      else m=m1->copy();
    }
    else
    {
      if ((m1->isA("Matrix"))&&(m2->isA("Constant")))
      {
	Matrix *mm1=(Matrix *)m1;
	Constant *cc2=(Constant *)m2;
	m=Matrix::pow(mm1,cc2,managem1,managem2);
      }
      else
      {
	Map *mm1=(managem1)?m1:m1->copy();
	Map *mm2=(managem2)?m2:m2->copy();
	m=new Operator('^', mm1, mm2);
      }
    }
    if (managem1) m1=0L;
    if (managem2) m2=0L;
  }
  return m;
}

Map *Map::simplify(void)
{
  if (!m_tree)
  {
    kdDebug() << "Trying to simplify a null tree" << endl;
    return 0L;
  }
  return m_tree->simplify();
}

void Map::simplifyFunction(void)
{
  if (!m_tree) kdDebug() << "Trying to simplify a null tree function (this will" 
    " cause a core dump, but this way I can check that this does really"
    " never happen :-)" << endl;
  kdDebug() << "Map::simpFunc" << endl;
  Map *m=m_tree->simplify();
  if (m)
  {
    delete m_tree;
    m_tree=m;
  }
}

void Map::propagateForceEvaluation(bool b)
{
  m_forceEvaluation=b;
  for (unsigned int i=0;i<m_nParameters;i++)
    if (m_parameters[i]->map()) 
      m_parameters[i]->map()->propagateForceEvaluation(b);

  if (m_tree) m_tree->propagateForceEvaluation(b);
}

Map *Map::add2(Map * m1, Map * m2, bool managem1, bool managem2)
{
  return Map::add(m1,m2,managem1,managem2);
}

Map *Map::sub2(Map * m1, Map * m2, bool managem1, bool managem2)
{
  return Map::sub(m1,m2,managem1,managem2);
}

Map *Map::mul2(Map * m1, Map * m2, bool managem1, bool managem2)
{
  return Map::mul(m1,m2,managem1,managem2);
}

Map *Map::div2(Map * m1, Map * m2, bool managem1, bool managem2)
{
  return Map::div(m1,m2,managem1,managem2);
}

Map *Map::pow2(Map * m1, Map * m2, bool managem1, bool managem2)
{
  return Map::pow(m1,m2,managem1,managem2);
}

Map *Map::derive(const Variable &var, const QDict <Map> &vars)
{
  if (!m_tree) 
  { 
    kdError() << "Can't derive a null tree !" << endl;
    Variable &v=const_cast<Variable &>(var);
    return new Derive(copy(),static_cast<Variable *>(v.copy()));
  };
  Variable *v=new Variable(parameter(0)->name());

  // We must evaluate the parameters
  QDict <Map> *vars2=evalParams(vars);
  
  /**
   * When deriving a map of more than one variable, we derive by one of
   * them while evaluating the function in the rest of them, so we'll
   * take advantage of having all the parameters evaluated in vars2
   * and just take out the respective variable while deriving with
   * respect to it.
   */
  Map *mtemp=vars2->take(m_parameters[0]->name());
  
  Map *f=m_tree->derive(*v,*vars2);
  vars2->replace(m_parameters[0]->name(),mtemp);

  if (!f) return 0L;
  for (unsigned int j=0;j<m_nParameters;j++)
    f->setParameterExpression(j,m_parameters[j]->copy());

  Map *r;
  Map *r2;

  r=m_parameters[0]->map()->derive(var,vars);
  
  r=Map::mul(f,r,true,true);
  delete v;

  for (unsigned int i=1;i<m_nParameters;i++)
  {
    if (argument(i))
    {
      Variable *v=new Variable(parameter(i)->name());
      mtemp=vars2->take(m_parameters[i]->name());
      Map *f=m_tree->derive(*v,*vars2); // XXX
      vars2->replace(m_parameters[i]->name(),mtemp);
      if (!f) return 0L;
      for (unsigned int j=0;j<m_nParameters;j++)
	f->setParameterExpression(j,m_parameters[j]->copy());
      r2=m_parameters[i]->map()->derive(var,vars);
      r2=Map::mul(f,r2,true,true);
      delete v;

      r=Map::add(r,r2,true,true);
    }
  }

  delete vars2;
  return r;
}

QDict <Map>Map::maps(255);

void Map::debugMaps(const QDict <Map>&vars) const
{
  kdDebug() << "Maps dict ---" << endl;
  QDictIterator <Map> it(vars);
  while (it.current())
  {
    kdDebug() << QString("%1 -> %2").arg(it.currentKey()).arg(it.current()->m_name) << endl;
    QString tab("   ");
    it.current()->printDebug(tab);
    ++it;
  }
  kdDebug() << "--- Maps" << endl;
}
