/*  adams.cpp - 
    This file is part of Kalamaris
    Copyright (C) 2000  Antonio Larrosa Jimenez
    Kalamaris' homepage : http://devel-home.kde.org/~larrosa/kalamaris.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "adams.h"
#include "../parameter.h"
#include "../constant.h"
#include "../matrix.h"
#include "../tmatrix.h"
#include "../vartype.h"
#include <kdebug.h>
#include <kapp.h>
#include <kstddirs.h>
#include <qfile.h>
#include <unistd.h>

MethodAdams::MethodAdams(int adamsN) : MapMethod(QString("MAdams%1").arg(adamsN))
{
  m_nParameters=6;
  m_parameters=new Parameter *[m_nParameters];

  m_parameters[0]=new Parameter("f");
  m_parameters[0]->setDataType("Matrix");
  m_parameters[1]=new Parameter("y0");
  m_parameters[1]->setDataType("Matrix");
  m_parameters[2]=new Parameter("x0");
  m_parameters[2]->setDataType("Matrix");
  m_parameters[3]=new Parameter("b");
  m_parameters[3]->setDataType("Constant");
  m_parameters[4]=new Parameter("m");
  m_parameters[4]->setDataType("Constant");
  m_parameters[5]=new Parameter("order");
  m_parameters[5]->setDataType("Constant");

  m_adamsN = adamsN;
}

MethodAdams::MethodAdams(const MethodAdams &m) : MapMethod(m)
{
  m_adamsN = m.m_adamsN;
}

MethodAdams::~MethodAdams()
{
}
  
Map *MethodAdams::copy(void) const
{
  return new MethodAdams(*this);
}

TMatrix *MethodAdams::method(const TMatrix &f, const TMatrix &x, 
    const TMatrix &y, QDict <Map> &vars, const TMatrix &B, unsigned int k, 
    unsigned int n, unsigned int p, const T &h)
{
  QDict <Map> tmpvars(vars);
  TMatrix H(3,n);
  TMatrix *S=new TMatrix(n,1);
  if (m_adamsN==1)
    S->setCol(0,const_cast<TMatrix *>(&y),k);
  else
    S->setCol(0,const_cast<TMatrix *>(&y),k-p+1);

  Map *fj;
  Map *tm;
  unsigned int i,j;
  
  T &tmp=T::newTG(0.0);
  T &tmp2=T::newTG(0.0);
  for (i=k-p+1;i<=k;i++)
  {
//    vars.clear();
    vars.replace("x",x.at(i));
    for (j=0;j<n;j++) 
      vars.replace(QString("y%1").arg(j),y.at(j,i));

    tmp=B.tRef(i-(k-p+1))*h;

    for (j=0;j<n;j++)
    {
      fj=f.at(j);

      /**
       * Now we evaluate f (x  , y    , ..., y   )
       *                  j  i    0,i         n,i
       */ 

      tm=fj->eval(vars);

      tmp2=*tm->t()*tmp;
      delete tm;

      S->set(j,0,new Constant(T::add(S->t(j,0),&tmp2)));
    }
  }

  return S;
}

Map *MethodAdams::eval(const QDict <Map> &vars)
{
  kdDebug() << "Adams::eval" << endl;
  QDict <Map> *vars2=evalParams(vars);
  if (!checkParamTypes(*vars2)) return 0L; // XXX Should return an error !
  TMatrix f(*vars2->find("f")->tmatrix());
  const TMatrix y0(*vars2->find("y0")->tmatrix());
  const TMatrix x0(*vars2->find("x0")->tmatrix());
  T &b=*T::newT(*vars2->find("b")->t());
  unsigned int n=f.nRows();
  unsigned int m=(int) *vars2->find("m")->t();
  unsigned int p=(int) *vars2->find("order")->t();
  
  T &tol=*T::newT(*maps.find("_tolerance")->t());
  T &hfactor=*T::newT(*maps.find("_stepControlFactor")->t());
  int maxIter=(int) *maps.find("_maxIterations")->t();
  T &e=*T::newT(0.0);
  bool stepControl=(*maps.find("_stepControl")->t()==e)?false:true;
  
//  kdDebug() << QString("t %1 h %2 i %3").arg(tol.string()).arg(hfactor.string()).arg(maxIter) << endl;
  
  QDict <Map> tmpvars(vars);

  TMatrix x(x0);
  TMatrix y(y0);
  TMatrix z(n,1);
  TMatrix H(5,n);
  unsigned int i,k;
  for (i=0;i<n;i++) 
  {
    f.at(i)->propagateForceEvaluation();
  }

  TMatrix B(1,1);
  B.load(locate("appdata",QString("coefficients/adams%1.%2").arg((m_adamsN==1)? "" : "2" ).arg(p)));
  
  T &h=*T::newT((b-x0.tRef(0))/(T::newTG(m)-T::newTG(1)));  // Seems a nice first step size
//  T &h=*T::newT((b-a)/(T::newTG(100)));  // Seems a nice first step size

  T &x2=*T::newT(0.0);
  T &x3=*T::newT(0.0);
  T &x4=*T::newT(0.0);
  T &x5=*T::newT(0.0);
  T &x6=*T::newT(0.0);

  TMatrix fx1(n,1);
  TMatrix fx2(n,1);
  TMatrix fx3(n,1);
  TMatrix fx4(n,1);
  TMatrix fx5(n,1);
  bool increaseH=false;

  T &two=*T::newT(2.0);
  
  // temptol = abs( (1/2)^4-1 )
//  T &temptol=*T::newT(0.9375);
  T &temptol=*T::newT(16.0/15);

  temptol*=tol;

  k=p-1;
  TMatrix *ytemp0=0L;
  TMatrix *ytemp1=0L;

  kdDebug() << "AAAA" << endl;
  while (x.tRef(k)<b)
  {
    kdDebug() << "k ( " << k << " )" << endl;
    kdDebug() << "xk ( " << static_cast<double>(x.tRef(k)) << " )" << endl;

    bool done=false;
    int it=0;
    while (!done)
    {
      Garbage::push();

      ytemp0=method(f,x,y,tmpvars,B,k,n,p,h);

      if (stepControl)
      {
	ytemp1=method(f,x,y,tmpvars,B,k,n,p,h/two);

	e=TMatrix::distMaxCol(ytemp0,0,ytemp1,0);
	delete ytemp1;

	it++;
	increaseH=false;

//	kdDebug() << QString("tol, e %1 %2 %3 %4").arg(temptol.string()).arg(e.string()).arg(h.string()).arg((h/two).string()) << endl;
	if ((e<=temptol)||(it>=maxIter)) 
	{
	  if (it<maxIter) // == if (e<=tol)
	  {
	    kdDebug() << QString("h = %1").arg(h.string()) << endl;
	    increaseH=true;
	  }
	  else
	    kdDebug() << QString("Max iterations exceeded : h = %1").arg(h.string()) << endl;

	  done=true;

	}
	else 
	{
	  h*=(tol/e).sqrt(p); 
	  h*=T::newTG(0.9);
	  kdDebug() << QString("Let's do another iteration: now with h=%1").arg(h.string()) << endl;
	  delete ytemp0;
	}
      }
      else  // the step control is disabled
	done=true;

      Garbage::pop();
//      kapp->processEvents(200);
    }

    k++;
    x.set(0,k,new Constant(T::add(x.t(k-1),&h)));
    y.setCol(k,ytemp0,0,true);

    if (increaseH) h*=hfactor;
  }
  kdDebug() << "Adams::eval(end)" << endl;
  delete vars2;
  
  TMatrix *tr=new TMatrix(1,2);
  tr->set(0,0, new Matrix(x));
  tr->set(0,1, new Matrix(y));
  
  T::deleteT(b);
  T::deleteT(e);
  T::deleteT(x2);
  T::deleteT(x3);
  T::deleteT(x4);
  T::deleteT(x5);
  T::deleteT(x6);
  T::deleteT(hfactor);
  T::deleteT(tol);

  kdDebug() << "Adams::eval(end2)" << endl;
  return new Matrix(tr);
}

QString MethodAdams::string(void) const
{
  return "MethodAdams";
}

QString MethodAdams::help(void) const
{
  return helpPrefix()+"<H3><I>MAdams( F, y0, a, b, M )</I></H3><BR><BR>Solves a system of "
    "differential equations using the Adams method"+helpPostfix();
}
