/*  midpoint.cpp - 
    This file is part of Kalamaris
    Copyright (C) 2000  Antonio Larrosa Jimenez
    Kalamaris' homepage : http://devel-home.kde.org/~larrosa/kalamaris.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "midpoint.h"
#include "../parameter.h"
#include "../constant.h"
#include "../matrix.h"
#include "../tmatrix.h"
#include "../vartype.h"
#include <kdebug.h>

MethodMiddlePoint::MethodMiddlePoint(void) : MapMethod("MMiddlePoint")
{
  m_nParameters=5;
  m_parameters=new Parameter *[m_nParameters];

  m_parameters[0]=new Parameter("f");
  m_parameters[0]->setDataType("Matrix");
  m_parameters[1]=new Parameter("y0");
  m_parameters[1]->setDataType("Matrix");
  m_parameters[2]=new Parameter("a");
  m_parameters[2]->setDataType("Constant");
  m_parameters[3]=new Parameter("b");
  m_parameters[3]->setDataType("Constant");
  m_parameters[4]=new Parameter("m");
  m_parameters[4]->setDataType("Constant");
}

MethodMiddlePoint::MethodMiddlePoint(const MethodMiddlePoint &m) : MapMethod(m)
{
}

MethodMiddlePoint::~MethodMiddlePoint()
{
}
  
Map *MethodMiddlePoint::copy(void) const
{
  return new MethodMiddlePoint(*this);
}

TMatrix *MethodMiddlePoint::method(const TMatrix &f,const TMatrix &x, const TMatrix &y, 
    QDict <Map> &vars,int k, int n,const T &h)
{
  Map *d, *fi;
  int i,j;
  TMatrix *yk=new TMatrix(n,1);
  T &t=*T::newT(0.0);
  T &t2=*T::newT(0.0);
  T &h2=*T::newT(h/T::newTG(2.0));
  
  for (i=0;i<n;i++)
  {
    fi=f.at(i);
    /**
     * Now we evaluate f (x   + h/2, y     + h/2 * d, ..., y     + h/2 * d )
     *                  i  k-1        0,k-1                 n,k-1
     *
     *  where d = f (x   , y     , ..., y     )
     *             i  k-1   0,k-1        n,k-1
     *
     * Note: fi is f
     *              i
     */ 
    vars.clear();

    vars.replace("x",x.at(k-1));
    for (j=0;j<n;j++) 
	vars.replace(QString("y%1").arg(j),y.at(j,k-1));

    d=fi->eval(vars);

    if (!d->isA("Constant")) 
      kdDebug() << "-----*************************************** NOT CONSTANT !!!" << endl;
    // We've calculated d, let's calculate the complex evaluation mentioned above .

    t=x.tRef(k-1)+h2;

    vars.replace("x",Constant::newConstantG(t));

    t2=(*d->t())*h2;

    for (j=0;j<n;j++) 
    {
      t=y.tRef(j,k-1)+t2;
      vars.replace(QString("y%1").arg(j),Constant::newConstantG(t));
    }

    d=fi->eval(vars);

    // Done, it's in d now

    //      delete d;   XXX
    yk->set(i,0,new Constant(y.tRef(i,k-1)+h*(*d->t())));
  }

  T::deleteT(h2);
  T::deleteT(t2);
  T::deleteT(t);
  return yk;
}
  
Map *MethodMiddlePoint::eval(const QDict <Map> &vars)
{
  kdDebug() << "MiddlePoint::eval" << endl;
  QDict <Map> *vars2=evalParams(vars);
  if (!checkParamTypes(*vars2)) return 0L; //XXX

  TMatrix f(*vars2->find("f")->tmatrix());
  const TMatrix y0(*vars2->find("y0")->tmatrix());
  T &a=*T::newT(*vars2->find("a")->t());
  T &b=*T::newT(*vars2->find("b")->t());
  unsigned int n=f.nRows();
  int m=(int) *vars2->find("m")->t();
  
  T &tol=*T::newT(*maps.find("_tolerance")->t());
  T &maxIterations=*T::newT(*maps.find("_maxIterations")->t());
  T &hfactor=*T::newT(*maps.find("_stepControlFactor")->t());
  T &e=*T::newT(0.0);
  bool stepControl=(*maps.find("_stepControl")->t()==e)?false:true;
  
  T &temptol=*T::newT(4.0/3);
  temptol*=tol;
  
  QDict <Map> tmpvars;

  TMatrix x(1,2);
  TMatrix y(n,2);
  TMatrix *ytemp0;
  TMatrix *ytemp1;
  unsigned int i,k;
  for (i=0;i<n;i++) 
  {
    y.set(i,0,y0.at(i)->copy());
    f.at(i)->propagateForceEvaluation();
  }
  
  T &hk=*T::newT((b-a)/(T::newTG(m)-T::newTG(1)));
  bool increaseH=false;

  x.set(0,0,new Constant(a));

  k=0;
  while (x.tRef(k)<b)
  {
    kdDebug() << "k ( " << k << " )" << endl;
    kdDebug() << "xk ( " << static_cast<double>(x.tRef(k)) << " )" << endl;
    k++;
    bool done=false;
    int it=0;
    while (!done)
    {
      x.set(0,k,new Constant(x.tRef(k-1)+hk));
      ytemp0=method(f,x,y,tmpvars,k,n,hk);

      if (stepControl)
      { 
	ytemp1=method(f,x,y,tmpvars,k,n,hk/T::newTG(2.0));
	
	e=*TMatrix::distMax(ytemp0,ytemp1);
	
	it++;
	increaseH=false;
	
	kdDebug() << QString("e %1").arg(e.string()) << endl;
	if ((e<=temptol)||(it>=maxIterations)) 
	{
	  if (it<maxIterations) // == if (e<=tol)
	  {
	    kdDebug() << QString("hk = %1").arg(hk.string()) << endl;
	    increaseH=true;
	  }
	  else
	    kdDebug() << QString("Max iterations exceeded : hk = %1").arg(hk.string()) << endl;
	  done=true;
	}
	else 
	{
	  kdDebug() << "Let's do another iteration with a smaller h" << endl;
	  hk*=((tol/e).sqrt(2));
	  hk*=T::newTG(0.9);
	  delete ytemp0;
	}
	delete ytemp1;
      }
      else  // the step control is disabled
	done=true;
    }

    y.setCol(k,ytemp0,0,true);

    if (increaseH) hk*=hfactor;
    
  }

  kdDebug() << "MiddlePoint::eval(end)" << endl;
  delete vars2;
  
  TMatrix *tr=new TMatrix(1,2);
  tr->set(0,0, new Matrix(x));
  tr->set(0,1, new Matrix(y));
  kdDebug() << "MiddlePoint::eval(end)2" << endl;
  
  return new Matrix(tr);
}

QString MethodMiddlePoint::string(void) const
{
  return "MethodMiddlePoint";
}

QString MethodMiddlePoint::help(void) const
{
  return helpPrefix()+"<H3><I>MMiddlePoint( F, y0, a, b, M )</I></H3><BR><BR>Solves a system of "
    "differential equations through the middle point method"+helpPostfix();
}
