/*  rk38.cpp - 
    This file is part of Kalamaris
    Copyright (C) 2000  Antonio Larrosa Jimenez
    Kalamaris' homepage : http://devel-home.kde.org/~larrosa/kalamaris.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "rk38.h"
#include "../parameter.h"
#include "../constant.h"
#include "../matrix.h"
#include "../tmatrix.h"
#include "../vartype.h"
#include <kdebug.h>
#include <kapp.h>
#include <unistd.h>

MethodRungeKutta38::MethodRungeKutta38(void) : MapMethod("MRungeKutta38")
{
  m_nParameters=5;
  m_parameters=new Parameter *[m_nParameters];

  m_parameters[0]=new Parameter("f");
  m_parameters[0]->setDataType("Matrix");
  m_parameters[1]=new Parameter("y0");
  m_parameters[1]->setDataType("Matrix");
  m_parameters[2]=new Parameter("a");
  m_parameters[2]->setDataType("Constant");
  m_parameters[3]=new Parameter("b");
  m_parameters[3]->setDataType("Constant");
  m_parameters[4]=new Parameter("m");
  m_parameters[4]->setDataType("Constant");
}

MethodRungeKutta38::MethodRungeKutta38(const MethodRungeKutta38 &m) : MapMethod(m)
{
}

MethodRungeKutta38::~MethodRungeKutta38()
{
}
  
Map *MethodRungeKutta38::copy(void) const
{
  return new MethodRungeKutta38(*this);
}

TMatrix *MethodRungeKutta38::method(const TMatrix &f,const T &x, const TMatrix &y,
        QDict <Map> &vars, unsigned int k, unsigned int n,const T &h)
{
  QDict <Map> tmpvars(vars);
  TMatrix H(3,n);
  
  Map *fi;
  unsigned int i,j;
  
  T &x1=*T::newT(x+T::newTG(1.0/3)*h); // Could use T::add to optimize it a bit
  T &x2=*T::newT(x+T::newTG(2.0/3)*h);
  T &x3=*T::newT(x+h);

  TMatrix fx0(n,1);
  TMatrix fx1(n,1);
  TMatrix fx2(n,1);
  TMatrix fx3(n,1);

  TMatrix *ry=new TMatrix(n,1);

  calcH0(f,y,tmpvars,k,n,h,H,x,x1,x2,fx0,fx1,fx2);
  calcH1(f,y,tmpvars,k,n,h,H,x,x1,x2,fx0,fx1,fx2);
  calcH2(f,y,tmpvars,k,n,h,H,x,x1,x2,fx0,fx1,fx2);

  T &oneEighth=T::newTG(1.0/8);
  T &threeEighth=T::newTG(3.0/8);
  T &t=T::newTG(0.0);
  for (i=0;i<n;i++)
  {
    fi=f.at(i);
    
    /**
     * Now we evaluate f (x   , y     , ..., y     )
     *                  i  k-1   0,k-1        n,k-1
     */ 

    vars.replace("x",Constant::newConstantG(x3));
    for (j=0;j<n;j++) 
	vars.replace(QString("y%1").arg(j),H.at(2,j));
    
    fx3.set(i,0,fi->eval(vars));

    t=h*( oneEighth*fx0.tRef(i)+threeEighth*fx1.tRef(i)+threeEighth*fx2.tRef(i)+oneEighth*fx3.tRef(i) );
    ry->set(i,0,new Constant(T::add(y.t(i,k),&t)));
  }

  return ry;
}

void MethodRungeKutta38::calcH0(const TMatrix &f, const TMatrix &y, 
    QDict <Map> &vars, unsigned int k, unsigned int n, const T &h, TMatrix &H,
    const T &x0, const T &, const T &,
    TMatrix &fx0, TMatrix &, TMatrix &)
{
  Map *fi;
  unsigned int i,j;

  for (i=0;i<n;i++)
  {
    fi=f.at(i);

    /**
     * Now we evaluate f (x   , y     , ..., y     )
     *                  i  k-1   0,k-1        n,k-1
     */ 

    vars.replace("x",Constant::newConstantG(x0));
    for (j=0;j<n;j++) 
      vars.replace(QString("y%1").arg(j),y.at(j,k));

    fx0.set(i,0,fi->eval(vars));

    H.set(0,i,new Constant(T::add(y.t(i,k),&(h*T::newTG(1.0/3)*fx0.tRef(i)))));
  }
}

void MethodRungeKutta38::calcH1(const TMatrix &f, const TMatrix &y, 
    QDict <Map> &vars, unsigned int k, unsigned int n,const T &h, TMatrix &H,
    const T &, const T &x1, const T &,
    TMatrix &fx0, TMatrix &fx1, TMatrix &)
{
  Map *fi;
  unsigned int i,j;
  
  T &c1=T::newTG(-1.0/3);
  T &t=T::newTG(0.0);
  for (i=0;i<n;i++)
  {
    fi=f.at(i);
    
    /**
     * Now we evaluate f (x   , y     , ..., y     )
     *                  i  k-1   0,k-1        n,k-1
     */ 

    vars.replace("x",Constant::newConstantG(x1));
    for (j=0;j<n;j++) 
      vars.replace(QString("y%1").arg(j),H.at(0,j));

    fx1.set(i,0,fi->eval(vars));

    t=h*(c1*fx0.tRef(i)+fx1.tRef(i) );
    H.set(1,i,new Constant(T::add(y.t(i,k),&t)));
  }

}

void MethodRungeKutta38::calcH2(const TMatrix &f, const TMatrix &y, 
    QDict <Map> &vars, unsigned int k, unsigned int n, const T &h, TMatrix &H,
    const T &, const T &, const T &x2,
    TMatrix &fx0, TMatrix &fx1, TMatrix &fx2)
{
  Map *fi;
  unsigned int i,j;
  T &t=T::newTG(0.0);

  for (i=0;i<n;i++)
  {
    fi=f.at(i);
    
    /**
     * Now we evaluate f (x   , y     , ..., y     )
     *                  i  k-1   0,k-1        n,k-1
     */ 

    vars.replace("x",Constant::newConstantG(x2));
    for (j=0;j<n;j++) 
      vars.replace(QString("y%1").arg(j),H.at(1,j));

    fx2.set(i,0,fi->eval(vars));

    t=h*( fx0.tRef(i)-fx1.tRef(i)+fx2.tRef(i) );
    H.set(2,i,new Constant(T::add(y.t(i,k),&t)));
  }

}

Map *MethodRungeKutta38::eval(const QDict <Map> &vars)
{
  kdDebug() << "RungeKutta38::eval" << endl;
  QDict <Map> *vars2=evalParams(vars);
  if (!checkParamTypes(*vars2)) return 0L;

  TMatrix f(*vars2->find("f")->tmatrix());
  const TMatrix y0(*vars2->find("y0")->tmatrix());
  T &a=*T::newT(*vars2->find("a")->t());
  T &b=*T::newT(*vars2->find("b")->t());
  unsigned int n=f.nRows();
  int m=(int) *vars2->find("m")->t();
  
  T &tol=*T::newT(*maps.find("_tolerance")->t());
  T &hfactor=*T::newT(*maps.find("_stepControlFactor")->t());
  int maxIter=(int) *maps.find("_maxIterations")->t();
  T &e=*T::newT(0.0);
  bool stepControl=(*maps.find("_stepControl")->t()==e)?false:true;
  
  QDict <Map> tmpvars;

  TMatrix x(1,2);
  TMatrix y(n,2);
  TMatrix z(n,1);
  TMatrix H(5,n);

  unsigned int i,k;
  for (i=0;i<n;i++) 
  {
    y.set(i,0,y0.at(i)->copy());
    f.at(i)->propagateForceEvaluation();
  }
  x.set(0,0,new Constant(a));
  
  T &h=*T::newT((b-a)/(T::newTG(m)-T::newTG(1)));  // Seems a nice first step size
//  T &h=*T::newT((b-a)/(T::newTG(100)));  // Seems a nice first step size

  T &x2=*T::newT(0.0);
  T &x3=*T::newT(0.0);
  T &x4=*T::newT(0.0);
  T &x5=*T::newT(0.0);
  T &x6=*T::newT(0.0);

  TMatrix fx1(n,1);
  TMatrix fx2(n,1);
  TMatrix fx3(n,1);
  TMatrix fx4(n,1);
  TMatrix fx5(n,1);
  bool increaseH=false;

  T &two=*T::newT(2.0);
  
  // temptol = abs( (1/2)^4-1 )
//  T &temptol=*T::newT(0.9375);
  T &temptol=*T::newT(16.0/15);

  temptol*=tol;
  kdDebug() << "RungeKutta38::eval3\n" << endl;

  k=0;
  TMatrix *ytemp0=0L;
  TMatrix *ytemp1=0L;
  while (x.tRef(k)<b)
  {
    kdDebug() << "k ( " << k << " )" << endl;
    kdDebug() << "xk ( " << static_cast<double>(x.tRef(k)) << " )" << endl;

    bool done=false;
    int it=0;
    while (!done)
    {
      Garbage::push();

      ytemp0=method(f,x.tRef(k),y,tmpvars,k,n,h);

      if (stepControl)
      {
	ytemp1=method(f,x.tRef(k),y,tmpvars,k,n,h/two);

	e=TMatrix::distMaxCol(ytemp0,0,ytemp1,0);
	delete ytemp1;

	it++;
	increaseH=false;

//	kdDebug() << QString("tol, e %1 %2 %3 %4").arg(temptol.string()).arg(e.string()).arg(h.string()).arg((h/two).string()) << endl;
	if ((e<=temptol)||(it>=maxIter)) 
	{
	  if (it<maxIter) // == if (e<=tol)
	  {
	    kdDebug() << QString("h = %1").arg(h.string()) << endl;
	    increaseH=true;
	  }
	  else
	    kdDebug() << QString("Max iterations exceeded : h = %1").arg(h.string()) << endl;

	  done=true;

	}
	else 
	{
	  h*=((tol/e).sqrt(4)); 
	  h*=T::newTG(0.9);
	  kdDebug() << QString("Let's do another iteration: now with h=%1").arg(h.string()) << endl;
	  delete ytemp0;
	}
      }
      else  // the step control is disabled
	done=true;

      Garbage::pop();
//      kapp->processEvents(200);
    }

  k++;
  x.set(0,k,new Constant(T::add(x.t(k-1),&h)));
  y.setCol(k,ytemp0,0,true);

  if (increaseH) h*=hfactor;

  }
  kdDebug() << "RungeKutta38::eval(end)" << endl;
  delete vars2;
  
  TMatrix *tr=new TMatrix(1,2);
  tr->set(0,0, new Matrix(x));
  tr->set(0,1, new Matrix(y));
  
  T::deleteT(a);
  T::deleteT(b);
  T::deleteT(e);
  T::deleteT(x2);
  T::deleteT(x3);
  T::deleteT(x4);
  T::deleteT(x5);
  T::deleteT(x6);
  T::deleteT(hfactor);
  T::deleteT(tol);

  kdDebug() << "RungeKutta38::eval(end2)" << endl;
  return new Matrix(tr);
}

QString MethodRungeKutta38::string(void) const
{
  return "MethodRungeKutta38";
}

QString MethodRungeKutta38::help(void) const
{
  return helpPrefix()+"<H3><I>MRungeKutta38( F, y0, a, b, M )</I></H3><BR><BR>Solves a system of "
    "differential equations using the Runge-Kutta 3/8 method (4th order)"+helpPostfix();
}
