/*  operator.cpp - 
    This file is part of Kalamaris
    Copyright (C) 2000  Antonio Larrosa Jimenez
    Kalamaris' homepage : http://devel-home.kde.org/~larrosa/kalamaris.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "operator.h"
#include "constant.h"
#include "matrix.h"
#include <stdio.h>
#include <kdebug.h>

Operator::Operator(char op, const QString &d1, const QString &d2) : Map()
{
  m_isAClassName="Operator";
  m_op=op;
  m_tree1=m_tree2=0L;
  if (!d1.isEmpty()) m_tree1=Map::parse(d1);
  if (!d2.isEmpty()) m_tree2=Map::parse(d2);
}

Operator::Operator(char op, Map *m1, Map *m2) : Map()
{
  m_isAClassName="Operator";
  m_op=op;
  m_tree1=m1;
  m_tree2=m2;
}

Operator::~Operator()
{
  delete m_tree1;
  delete m_tree2;
}
  

Map *Operator::copy(void) const
{
  return new Operator(m_op,m_tree1->copy(),m_tree2->copy());
}

Map *Operator::eval (const QDict <Map> &vars)
{
//  kdDebug() << "Operator::eval" << endl;
  Map *r1=m_tree1->eval(vars);
  Map *r2=m_tree2->eval(vars);
  Map *r=0L;

  switch (m_op)
    {  
      case '+' : r=Map::add(r1,r2,true,true);break;
      case '-' : r=Map::sub(r1,r2,true,true);break;
      case '*' : r=Map::mul(r1,r2,true,true);break;
      case '/' : r=Map::div(r1,r2,true,true);break;
      case '^' : r=Map::pow(r1,r2,true,true);break;
		 //case '^' : return pow(tree1->eval(vars),tree2->eval(vars));break;
      default : fprintf(stderr,"Unknown Operator %c !!\n",m_op); break;
    }
 
//  printf("operator %c\n",op);
//  kdDebug() << r1->string() << endl;
//  kdDebug() << r2->string() << endl;
//  kdDebug() << r->string() << endl;
  
//  delete r1;
//  delete r2;
//  printf("--\n");
  return r;
}

QString Operator::string(void) const
{
  if ((m_op=='*')||(m_op=='/')||(m_op=='^'))
  {
    QString s;
    bool writeOperator=true;
    if ( m_tree1->isA("Operator") && 
	(reinterpret_cast<Operator *>(m_tree1)->m_op=='+' || 
	 reinterpret_cast<Operator *>(m_tree1)->m_op=='-' ) )
      s+=QString("(%1)").arg(m_tree1->string());
    else 
      if ((m_tree1->isA("Constant"))&&(m_op=='*'))
      {
	if (*m_tree1->t()==T::newTG(-1.0)) s+="-";
	else s+=m_tree1->string();
	writeOperator=false;
      }
      else
      s+=m_tree1->string();
    
    if (writeOperator) s+=m_op;
    
    if ( m_tree2->isA("Operator") &&
	(reinterpret_cast<Operator *>(m_tree2)->m_op=='+' || 
	 reinterpret_cast<Operator *>(m_tree2)->m_op=='-' ) )
      s+=QString("(%1)").arg(m_tree2->string());
    else
      s+=m_tree2->string();
    
    return s;
  }
  else
    return m_tree1->string()+m_op+m_tree2->string();
}

void Operator::propagateForceEvaluation(bool b)
{
  m_forceEvaluation=b;
  if (m_tree1) m_tree1->propagateForceEvaluation(b);
  if (m_tree2) m_tree2->propagateForceEvaluation(b);
}

Map *Operator::simplify(void)
{
  if ((!m_tree1)||(!m_tree2)) return 0L; //XXX
  if ((m_tree1->isA("Constant"))&&(m_tree2->isA("Constant")))
  {
      QDict <Map> emptydict;
      return eval(emptydict);
  }
  /**
   * Should also try to simplify trees on the form:
   *
   *    
   *    +
   *   / \
   *  1   -
   *     / \
   *    x   1
   *
   * Or even in the form:
   *
   *         -
   *        / \
   *       /   \
   *      /     \
   *     *       *
   *    / \     / \
   *   x   y   y   x
   *
   * (If we're in a commutative ring :-) )
   */
  Map *s1=m_tree1->simplify();
  Map *s2=m_tree2->simplify();
  if ((!s1)&&(!s2)) return 0L;
  kdDebug() << "Operator::Simplifying !!!" << endl;
//  kdDebug() << QString("%1 %2 %3 %4").arg(tree1->string()).arg(s1->string())//      .arg(tree2->string()).arg(s2->string()) << endl;

  // What happens if s1 is simplified but s2 wasn't ?!?!
  if (!s1) s1=m_tree1->copy();
  if (!s2) s2=m_tree2->copy();

  return new Operator(m_op,s1,s2);
}


Map *Operator::derive(const Variable &var, const QDict <Map> &vars)
{
  kdDebug() << "Op::derive" << endl;
  Map *r1=m_tree1->derive(var, vars);
  Map *r2=m_tree2->derive(var, vars);
  Map *r=0L;
  
  switch (m_op)
    {  
      case '+' : 
	{
	  r=Map::add(r1,r2,true,true);
	} break;
      case '-' : 
	{
	  r=Map::sub(r1,r2,true,true);
	} break;
      case '*' : 
	{
	  Map *t1=Map::mul(r1,m_tree2,true,false);
	  Map *t2=Map::mul(m_tree1,r2,false,true);
	  r=Map::add(t1,t2,true,true);
	} break;
      case '/' : 
	{
	  Map *t1=Map::mul(r1,m_tree2,true,false);
	  Map *t2=Map::mul(m_tree1,r2,false,false);
	  Map *p1=Map::sub(t1,t2,true,true);
	  Map *p2=Map::pow2(m_tree2,new Constant(2.0),false,true);
	  r=Map::div(p1,p2,true,true);
	} break;
      case '^' : 
	{
	  if (r2->isA("Constant")) 
	  {
	    Map *e=Map::sub2(m_tree2,new Constant(1.0),false,true);
	    Map *p=Map::pow(m_tree1,e,false,true);
	    r=Map::mul(m_tree2,p,false,true);
	  }
	} break;
      default : fprintf(stderr,"Op::derive : Unknown Operator %c !!\n",m_op); break;
    }
  return r;
}
