/*  tmatrix.cpp - TMatrix class
    This file is part of Kalamaris
    Copyright (C) 2000  Antonio Larrosa Jimenez
    Kalamaris' homepage : http://devel-home.kde.org/~larrosa/kalamaris.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "matrix.h"
#include "map.h"
#include "constant.h"
#include "operator.h"
#include <kdebug.h>
#include <qfile.h>
#include <stdio.h>

TMatrix::TMatrix(unsigned int _rdim, unsigned int _cdim, Map ***_maps): T(0.0)
{
  nrows=_rdim;
  ncols=_cdim;
  maxrows=nrows;
  maxcols=ncols;
  matrix=_maps;
  m_autoExpand=true;
  if (matrix==0L)
    {
      matrix=new Map **[nrows];
      for (unsigned int i=0;i<nrows;i++)
      {
	matrix[i]=new Map *[ncols];
      for (unsigned int j=0;j<ncols;j++)
//        matrix[i][j]=new Constant(0.0);
        matrix[i][j]=0L;
      }
    }
}

TMatrix::TMatrix( const TMatrix &x) : T(0.0)
{
  ncols=x.ncols;
  nrows=x.nrows;
  maxrows=nrows;
  maxcols=ncols;
  m_autoExpand=true;
  matrix=new Map **[nrows];
  for (unsigned int i=0;i<nrows;i++)
  {
    matrix[i]=new Map *[ncols];
    for (unsigned int j=0;j<ncols;j++)
      if (x.matrix[i][j]) matrix[i][j]=x.matrix[i][j]->copy(); 
      else matrix[i][j]=0L;
  }
}

TMatrix::~TMatrix(void)
{
  for (unsigned int i=0;i<maxrows;i++)
  {
    for (unsigned int j=0;j<maxcols;j++) delete matrix[i][j];
    delete matrix[i];
  }
  delete matrix;
}

void TMatrix::clear(void)
{
  for (unsigned int i=0;i<maxrows;i++)
  {
    for (unsigned int j=0;j<maxcols;j++) 
    {
      delete matrix[i][j];
      matrix[i][j]=0L;
    }
  }
}

TMatrix *TMatrix::copy(void) const
{
  return new TMatrix(*this);  
}

void TMatrix::expandRows(unsigned int i)
{
  if (i>=maxrows)
  {
    maxrows=i+11;
    //Now maxrows is greater than i
    Map ***tmp=matrix;
    matrix=new Map **[maxrows];
    // First, we copy (reassign) the rows we already had
    for (unsigned int i=0;i<nrows;i++) matrix[i]=tmp[i];

    delete tmp;

    // Now, we initialize the rest of the rows (even if
    // we're not using them yet)
    for (unsigned int i=nrows;i<maxrows;i++)
    {
      matrix[i]=new Map *[ncols];
      for (unsigned int j=0;j<ncols;j++)
	matrix[i][j]=0L;
    }
  }
  nrows=i+1; 
}

void TMatrix::expandCols(unsigned int j)
{
  if (j>=maxcols)
  {
    kdDebug() << "expandCols " << ncols << " " << maxcols << endl;
    maxcols=j+11;
    //Now maxcols is greater than j
    Map **tmp;
    for (unsigned int i=0;i<maxrows;i++)
    {
      tmp=matrix[i];
      matrix[i]=new Map *[maxcols];
      for (unsigned int j=0;j<ncols;j++) matrix[i][j]=tmp[j];

      delete tmp;

      // Now, we initialize the rest of the cols (even if
      // we're not using them yet)
      for (unsigned int j=ncols;j<maxcols;j++)
      {
	matrix[i][j]=0L;
      }

    }
  }
  ncols=j+1;
}

void TMatrix::set(unsigned int i, unsigned int j, Map *_m)
{
//  kdDebug() << QString("TMatrix::set(%1,%2, %3)").arg(i).arg(j).arg(_m->isA()) << endl;
  if ((i>=nrows)||(j>=ncols)) 
  {
//    kdError() << QString("TMatrix::set (%1>=%2 || %3>=%4)").//			arg(i).arg(nrows).arg(j).arg(ncols) << endl;
    if (!isAutoExpandable()) return;

    if (i>=nrows) // We have to do something about the number of rows
      expandRows(i);

    if (j>=ncols) // We have to do something about the number of cols
      expandCols(j);

//    kdError() << QString("TMatrix::set (%1<%2 && %3<%4)").//			arg(i).arg(nrows).arg(j).arg(ncols) << endl;
  }
//  kdDebug() << QString("TMatrix::set") << endl;
  delete matrix[i][j];
//  matrix[i][j]=_m->copy();
  matrix[i][j]=_m;
//  kdDebug() << QString("$ %1 %2 (%3)").arg(i).arg(j).arg(matrix[i][j]->isA()) << endl;
}

Map *TMatrix::at(unsigned int i, unsigned int j) const
{
  if ((i>=nrows)||(j>=ncols)) 
  {
    kdError() << QString("TMatrix::at (%1>=%2 || %3>=%4)").			arg(i).arg(nrows).arg(j).arg(ncols) << endl;
    return 0L;
  }

  return matrix[i][j];
}

Map *TMatrix::at(unsigned int i) const
{
  if (nrows==1) 
  {
    if (i>=ncols) 
    {
      kdError() << QString("TMatrix::at (------ || %1>=%2)").	  arg(i).arg(ncols) << endl;
      return 0L;
    }
    return matrix[0][i];
  }
  if (i>=nrows)
  {
    kdError() << QString("TMatrix::at (%1>=%2 || ------)").	arg(i).arg(nrows) << endl;
    return 0L;
  }
  return matrix[i][0];
}

void TMatrix::setCol(unsigned int j, TMatrix *v, unsigned int vj, bool managev)
{
//  if (v->nRows()==1) v->traspose(); // We want it vertical
  //Not sure about the intentions of the above line when I wrote it,
  // but I don't think it will work if we don't manage v now
  // that this is a general method

  if (v->nRows()==1) 
  {
    kdError() << "TMatrix::setCol : It's not vertical !" << endl;
    return;
  }
  
  if (nrows!=v->nRows()) 
  {
    kdError() << "TMatrix::setCol : dimensions are different !" << endl;
    return;
  }

  if (j>=ncols) 
  {
    if (!isAutoExpandable())
    {
      kdError() << QString("TMatrix::setCol (------ || %1>=%2)").	arg(j).arg(ncols) << endl;
      return;
    }
    expandCols(j);
  }

  if (managev)
  {
    for (unsigned int i=0;i<nrows;i++)
    {
      delete matrix[i][j];
      matrix[i][j]=v->matrix[i][vj];
      v->matrix[i][vj]=0L;
    }
    delete v;
  }
  else
  {
    for (unsigned int i=0;i<nrows;i++) 
    {
      delete matrix[i][j];
      matrix[i][j]=v->matrix[i][vj]->copy();
    }
  }
}

void TMatrix::setRow(unsigned int i, TMatrix *v, unsigned int vi, bool managev) 
{
//  if (v->nCols()==1) v->traspose(); // We want it horizontal
  //Not sure about the intentions of the above line when I wrote it,
  // but I don't think it will work if we don't manage v now
  // that this is a general method

  if (v->nCols()==1) 
  {
    kdError() << "TMatrix::setCol : It's not vertical !" << endl;
    return;
  }
 
  if (ncols!=v->nCols()) 
  {
    kdError() << "TMatrix::setRow : dimensions are different !" << endl;
    return;
  }

  if (i>=nrows) 
  {
    if (!isAutoExpandable())
    {
      kdError() << QString("TMatrix::setRow (%1>=%2 || ------)").	arg(i).arg(nrows) << endl;
      return;
    }
    expandRows(i);
  }

  if (managev)
  {
    for (unsigned int j=0;j<ncols;j++)
    {
      delete matrix[i][j];
      matrix[i][j]=v->matrix[vi][j];
      v->matrix[vi][j]=0L;
    }
    delete v;
  }
  else
  {
    for (unsigned int j=0;j<ncols;j++) 
    {
      delete matrix[i][j];
      matrix[i][j]=v->matrix[vi][j]->copy();
    }
  }
}

inline int TMatrix::operator==( TMatrix ) const
{
  return 0;
  //  return v==v1.v;
}
  
inline int TMatrix::operator==( double ) const
{
  return 0;
 // return v==g;
}

inline int TMatrix::operator!=( TMatrix ) const
{
  return 0;
//  return v!=v1.v;
}

inline int TMatrix::operator!=( double ) const
{
  return 0;
  //return v!=g;
}

inline int TMatrix::operator<( TMatrix ) const
{
  return 0;
 // return v<v1.v;
}

inline int TMatrix::operator<( double ) const
{
  return 0;
//  return v<g;
}

inline int TMatrix::operator>( TMatrix ) const
{
  return 0;
//  return v>v1.v;
}

inline int TMatrix::operator>( double ) const
{
  return 0;
 // return v>g;
}

inline int TMatrix::operator>=( TMatrix ) const
{
  return 0;
//  return v>=v1.v;
}

inline int TMatrix::operator>=( double ) const
{
  return 0;
//  return v>=g;
}

inline int TMatrix::operator<=( TMatrix ) const
{
  return 0;
//  return v<=v1.v;
}

inline int TMatrix::operator<=( double ) const
{
  return 0;
//  return v<=g;
}

TMatrix &TMatrix::operator=( const TMatrix &x )
{ 
  ncols=x.ncols;
  nrows=x.nrows;

  // XXX XXX delete matrix
  matrix=new Map **[nrows];
  for (unsigned int i=0;i<nrows;i++)
  {
    matrix[i]=new Map *[ncols];
    for (unsigned int j=0;j<ncols;j++)
      matrix[i][j]=x.matrix[i][j]->copy(); 
  }
  return *this;
}

TMatrix &TMatrix::operator=( double & )
{
  ncols=1;
  nrows=1;
  // XXX XXX delete matrix
  matrix=new Map **[nrows];
  for (unsigned int i=0;i<nrows;i++)
  {
    matrix[i]=new Map *[ncols];
    for (unsigned int j=0;j<ncols;j++)
      matrix[i][j]=new Constant("0"); 
  }
  return *this;
}

/*
TMatrix *TMatrix::operator+( const T *_v1 )
{
  TMatrix *v1=(TMatrix *)_v1;
  printf("AAAAAAAAAAAAAAAAAAAAAAAA TMatrix::+ \n");  
  TMatrix *tmp=new TMatrix( nrows, ncols );
  for  (int i=0;i<nrows;i++)
  {
    for (int j=0;j<ncols;j++)
    {
//      kdDebug() << i << ", " << j << endl;
//      kdDebug() << matrix[i][j]->isA() << endl;
//      kdDebug() << v1->matrix[i][j]->isA() << endl;

      if ((matrix[i][j]->isA("Constant"))
	   &&(v1->matrix[i][j]->isA("Constant")))
      {
	printf("Optimiz\n");
	tmp->matrix[i][j]=new Constant(
	    *(((Constant *)matrix[i][j])->t()) +
	     *(((Constant *)v1->matrix[i][j])->t()) );
      }
      else
      if ((matrix[i][j]->isA("Matrix"))
	   &&(v1->matrix[i][j]->isA("Matrix")))
      {
	tmp->matrix[i][j]=new Matrix( (TMatrix *)&(
	    *(((Matrix *)matrix[i][j])->t()) +
	     *(((Matrix *)v1->matrix[i][j])->t())) );
	
      }
      else
      tmp->matrix[i][j]=new Operator('+', matrix[i][j]->copy(), v1->matrix[i][j]->copy());
      
    }
  }    
  return tmp;
}

TMatrix *TMatrix::operator-( const T *_v1 )
{
  TMatrix *v1=(TMatrix *)_v1;
  TMatrix *tmp=new TMatrix( nrows, ncols);
  for  (int i=0;i<nrows;i++)
  {
    for (int j=0;j<ncols;j++)
      tmp->matrix[i][j]=new Operator('-', matrix[i][j]->copy(), v1->matrix[i][j]->copy());
  }    
  return tmp;
}

TMatrix *TMatrix::operator*( const T *_v1 )
{
  TMatrix *v1=(TMatrix *)_v1;
  TMatrix *tmp=new TMatrix( nrows, v1->ncols );
  Map *temp,*temp2;
  temp2=new Constant("0"); 
  for  (int i=0;i<nrows;i++)
  {
    for (int j=0;j<v1->ncols;j++)
    {
      for (int k=0;k<ncols;k++)
      {
	temp=new Operator('*', matrix[i][k]->copy(), v1->matrix[k][j]->copy());
	temp2=new Operator('+', temp, temp2);
      }
	tmp->matrix[i][j]=temp2;
    }
  }    
  return tmp;
}

TMatrix *TMatrix::operator/( const T * )
{
  return 0L; //XXX
}
*/

TMatrix &TMatrix::operator+=( const TMatrix &x )
{
//  Map *m;
//  printf("TMatrix::+ \n");  
  for  (unsigned int i=0;i<nrows;i++)
  {
    for (unsigned int j=0;j<ncols;j++)
    {
      //      kdDebug() << i << ", " << j << endl;
      //      kdDebug() << matrix[i][j]->isA() << endl;
      //      kdDebug() << v1->matrix[i][j]->isA() << endl;

/*      if ((matrix[i][j]->isA("Constant"))
	  &&(x.matrix[i][j]->isA("Constant")))
      {
	printf("Optimiz\n");
	m=new Constant( (*(matrix[i][j])->t()) + (*(x.matrix[i][j])->t()) );
	delete matrix[i][j];
      }
      else
	if ((matrix[i][j]->isA("Matrix"))
	    &&(x.matrix[i][j]->isA("Matrix")))
	{
	  m=new Matrix( (TMatrix *)&(
		*(((Matrix *)matrix[i][j])->t()) +
		*(((Matrix *)x.matrix[i][j])->t())) );
	  delete matrix[i][j];
	}
	else
	{
	  m=new Operator('+', matrix[i][j], x.matrix[i][j]->copy());
	}
      matrix[i][j]=m;*/
      matrix[i][j]=Map::add(matrix[i][j],x.matrix[i][j],true,false);
    }
  } 
//  printf("TMatrix::+(end) \n");  
  return *this;
}
 
TMatrix *TMatrix::add(TMatrix *&m1, TMatrix *&m2, bool managem1, bool managem2)
{
  unsigned int nrows=m1->nRows();
  unsigned int ncols=m1->nCols();
  TMatrix *m=new TMatrix(nrows,ncols);
  for  (unsigned int i=0;i<nrows;i++)
    for (unsigned int j=0;j<ncols;j++)
      m->matrix[i][j]=Map::add(m1->matrix[i][j],m2->matrix[i][j],managem1,managem2);

  if (managem1) { delete m1; m1=0L; };
  if (managem2) { delete m2; m2=0L; };
  return m;
}

TMatrix &TMatrix::operator-=( const TMatrix &x )
{
  for  (unsigned int i=0;i<nrows;i++)
    for (unsigned int j=0;j<ncols;j++)
       matrix[i][j]=Map::sub(matrix[i][j],x.matrix[i][j],true,false);

  return *this;
}

TMatrix *TMatrix::sub(TMatrix *&m1, TMatrix *&m2, bool managem1, bool managem2)
{
  unsigned int nrows=m1->nRows();
  unsigned int ncols=m1->nCols();
  TMatrix *m=new TMatrix(nrows,ncols);
  for  (unsigned int i=0;i<nrows;i++)
    for (unsigned int j=0;j<ncols;j++)
      m->matrix[i][j]=Map::sub(m1->matrix[i][j],m2->matrix[i][j],managem1,managem2);

  if (managem1) { delete m1; m1=0L; };
  if (managem2) { delete m2; m2=0L; };
  return m;
}

TMatrix &TMatrix::operator*=( const TMatrix &x )
{
  for  (unsigned int i=0;i<nrows;i++)
    for (unsigned int j=0;j<ncols;j++)
       matrix[i][j]=Map::mul(matrix[i][j],x.matrix[i][j],true,false);

  return *this;
}

TMatrix *TMatrix::mul(TMatrix *&m1, TMatrix *&m2, bool managem1, bool managem2)
{
  unsigned int nrows=m1->nRows();
  unsigned int ncols=m1->nCols();
  TMatrix *m=new TMatrix(nrows,ncols);
  for  (unsigned int i=0;i<nrows;i++)
    for (unsigned int j=0;j<ncols;j++)
      m->matrix[i][j]=Map::mul(m1->matrix[i][j],m2->matrix[i][j],managem1,managem2);

  if (managem1) { delete m1; m1=0L; };
  if (managem2) { delete m2; m2=0L; };
  return m;
}

TMatrix *TMatrix::mul(T *&c1, TMatrix *&m2, bool managem1, bool managem2)
{
  unsigned int nrows=m2->nRows();
  unsigned int ncols=m2->nCols();
  TMatrix *m=new TMatrix(nrows,ncols);
  for  (unsigned int i=0;i<nrows;i++)
    for (unsigned int j=0;j<ncols;j++)
    {
      Map *cc1=(Map *)new Constant(*c1);
      m->matrix[i][j]=Map::mul(cc1,m2->matrix[i][j],true,managem2);
    }

  if (managem1) { delete c1; c1=0L; };
  if (managem2) { delete m2; m2=0L; };
  return m;
}

TMatrix &TMatrix::operator/=( const TMatrix &x )
{
  for  (unsigned int i=0;i<nrows;i++)
    for (unsigned int j=0;j<ncols;j++)
       matrix[i][j]=Map::div(matrix[i][j],x.matrix[i][j],true,false);

  return *this;
} 

TMatrix *TMatrix::div(TMatrix *&m1, TMatrix *&m2, bool managem1, bool managem2)
{
  unsigned int nrows=m1->nRows();
  unsigned int ncols=m1->nCols();
  TMatrix *m=new TMatrix(nrows,ncols);
  for  (unsigned int i=0;i<nrows;i++)
    for (unsigned int j=0;j<ncols;j++)
      m->matrix[i][j]=Map::div(m1->matrix[i][j],m2->matrix[i][j],managem1,managem2);

  if (managem1) { delete m1; m1=0L; };
  if (managem2) { delete m2; m2=0L; };
  return m;
}

TMatrix *TMatrix::pow(TMatrix *&m1, T *&c2, bool managem1, bool managem2)
{
/// XXX It should check for c2==0 or negative
  TMatrix *m=new TMatrix(*m1);
  int c=(int)*c2;
  for (;c>1;c--)
    m=TMatrix::mul(m1,m,false,true);

  if (managem1) { delete m1; m1=0L; };
  if (managem2) { delete c2; c2=0L; };
  return m;
}

QString TMatrix::string(void) const
{
    QString s("[ ");
    
    for (unsigned int i=0;i<nrows;i++)
    {
      for (unsigned int j=0;j<ncols;j++)
      {
	if (matrix[i][j]) s+=matrix[i][j]->string(); 
	else s+="?";
	if (j+1<ncols) s+=", ";
	else if (i+1<nrows) s+="; ";
      }
    }
    s+=" ]";
    return s;
} 

QString TMatrix::htmlString(void) const
{
    QString s("<TABLE CELLSPACING=0 CELLPADDING=0 BORDER=0><TR VALIGN=CENTER><TD>[</TD>");
    
    s+="<TD><TABLE CELLSPACING=5 CELLPADDING=5 BORDER=0>";
    for (unsigned int i=0;i<nrows;i++)
    {
      s+="<TR ALIGN=CENTER>";
      for (unsigned int j=0;j<ncols;j++)
      {
	s+="<TD>";
	if (matrix[i][j]) s+=matrix[i][j]->htmlString(); 
	else s+="?";
	s+="</TD>";
      }
      s+="</TR>";
    }
    s+="</TABLE></TD><TD>]</TD></TR></TABLE>";
    return s;
}

void TMatrix::traspose(void)
{
  Map ***matback=matrix;
 // XXX This is buggy, isn't it ? (should do a semi-deep copy
  unsigned int i,j,t;
  t=nrows;
  nrows=ncols;
  ncols=t;

  matrix=new Map **[nrows];
  for (i=0;i<nrows;i++)
  {
    matrix[i]=new Map *[ncols];
    for (j=0;j<ncols;j++)
      matrix[i][j]=matback[j][i];
  }
  for (i=0;i<ncols;i++) // ncols are matback's nrows
  {
    delete matback[i];
  }
  delete matback;
  
}

void TMatrix::setAutoExpand(bool b)
{
  m_autoExpand=b;
}
  
T *TMatrix::distMax(TMatrix *v,TMatrix *w) 
{
  if ((v->nRows()!=w->nRows())||(v->nCols()!=w->nCols())) return T::newT(0.0);

  unsigned int i,j;
  T *dist=T::newT(0.0);
  T &tmp=*T::newT(0.0);

  for (i=0;i<v->nRows();i++)
    for (j=0;j<v->nCols();j++)
    {
      tmp=(*v->matrix[i][j]->t())-(*w->matrix[i][j]->t());
      tmp.abs();
//      *dist=QMAX(*dist,tmp);
      if (tmp>*dist) *dist=tmp;
    }
  T::deleteT(tmp);
  
  return dist;
}

T &TMatrix::distMaxCol(TMatrix *v, unsigned int c1, TMatrix *w, unsigned int c2) 
{
  if (v->nRows()!=w->nRows()) { kdError() << "TMatrix::distMaxCol : " 
    "Matrices don't have the same number of rows !" << endl; return T::newTG(0.0);}

  T *dist=T::newT(0.0);
  Garbage::throwT(dist);
  
  T &tmp=*T::newT(0.0);
  for (unsigned int i=0;i<v->nRows();i++)
    {
      tmp=(*v->matrix[i][c1]->t())-(*w->matrix[i][c2]->t());
      tmp.abs();
      if (tmp>*dist) *dist=tmp;
    }
  T::deleteT(tmp);
  
  return *dist;
}

T *TMatrix::t(unsigned int i) const 
{ 
  return at(i)->t(); 
}

T *TMatrix::t(unsigned int i, unsigned int j) const 
{ 
  return at(i,j)->t(); 
}

T &TMatrix::tRef(unsigned int i) const 
{ 
  return *at(i)->t(); 
}

T &TMatrix::tRef(unsigned int i,unsigned int j) const 
{ 
  return *at(i,j)->t(); 
}

void TMatrix::load(const QString &filename)
{
  bool backautoexpand=isAutoExpandable();
  setAutoExpand(true);
  
  clear();


  QFile f(filename);
  f.open( IO_ReadOnly );
  unsigned int i=0;

  kdDebug() << QString("TMatrix::load (%1)").arg(filename) << endl;
  while (!f.atEnd())
  {
    QString line;
    if (f.readLine(line,32767)!=-1)
    {
      QStringList elements=QStringList::split(" ",line.left(line.length()-1));
      for (unsigned int j=0;j<elements.count();j++)
	set( i, j, Map::parse(elements[j]));
    }
    i++;
  }

  setAutoExpand(backautoexpand);
}

TMatrix *TMatrix::submatrix(unsigned int i0,unsigned int j0, unsigned i1, unsigned int j1) const
{
  if (i0>=nrows) return 0L;
  if (i1>=nrows) return 0L;
  if (j0>=ncols) return 0L;
  if (j1>=ncols) return 0L;
  if (i0 > i1) return 0L;
  if (j0 > j1) return 0L;

  TMatrix *m=new TMatrix(i1-i0+1,j1-j0+1);
  
  for (unsigned int i=i0; i<=i1; i++)
    for (unsigned int j=j0; j<=j1; j++)
      m->matrix[i-i0][j-j0]=matrix[i][j]->copy();

  return m;
} 

TMatrix *TMatrix::column(unsigned int j) const
{
  return submatrix(0,j,nrows-1,j);
}

TMatrix *TMatrix::row(unsigned int i) const
{
  return submatrix(i,0,i,ncols-1);
}
